"""Tests for the BitfountAM and BitfountHub API calls."""
from datetime import datetime, timezone
import inspect
import json
import logging
from pathlib import Path
import re
from typing import List, Optional, Type, cast
from unittest.mock import Mock, PropertyMock

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
import pytest
from pytest import fixture
from pytest_lazyfixture import lazy_fixture
from pytest_mock import MockerFixture
import requests
from requests import HTTPError, RequestException
from requests.exceptions import InvalidJSONError
import responses
from responses.matchers import json_params_matcher

from bitfount import web_utils
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.types import SerializedProtocol, _PodResponseType
from bitfount.hub.api import (
    _DEV_AM_URL,
    _MAX_CUSTOM_MODEL_SIZE_BYTES,
    _MAX_CUSTOM_MODEL_SIZE_MEGABYTES,
    _MAX_SCHEMA_SIZE_BYTES,
    _MAX_SCHEMA_SIZE_MEGABYTES,
    _STAGING_AM_URL,
    PRODUCTION_AM_URL,
    BitfountAM,
    BitfountHub,
    PodPublicMetadata,
    _check_pod_id_details,
    hash_file_contents,
)
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.hub.exceptions import ModelUploadError, SchemaUploadError
from bitfount.hub.types import (
    _AccessManagerKeyResponseJSON,
    _AMAccessCheckResponseJSON,
    _HubFailureResponseJSON,
    _HubSuccessResponseJSON,
    _ModelDetailsResponseJSON,
    _ModelUploadResponseJSON,
    _MonitorPostJSON,
    _MultiModelDetailsResponseJSON,
    _MultiPodDetailsResponseJSON,
    _OIDCAccessCheckPostJSON,
    _PodDetailsResponseJSON,
    _SAMLAdditionalInfoPOSTJSON,
    _SAMLChallengeResponseJSON,
    _SignatureBasedAccessCheckPostJSON,
)
from bitfount.models.bitfount_model import BitfountModel
from bitfount.storage import _get_packed_data_object_size
from bitfount.types import (
    _JSONDict,
    _S3PresignedPOSTFields,
    _S3PresignedPOSTURL,
    _S3PresignedURL,
    _SAMLResponse,
)
from bitfount.utils import _get_mb_from_bytes
from tests.utils.helper import (
    get_debug_logs,
    get_error_logs,
    get_warning_logs,
    integration_test,
    unit_test,
)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950


def sign_message(message: bytes, private_key: RSAPrivateKey) -> bytes:
    """Signs provided `message` with provided `private_key` and returns signature."""
    signature = private_key.sign(
        message,
        padding.PSS(
            mgf=padding.MGF1(hashes.SHA256()), salt_length=20  # padding.PSS.MAX_LENGTH
        ),
        hashes.SHA256(),
    )

    return signature


@fixture
def pod_namespace() -> str:
    """Pod namespace information."""
    return "pod_namespace"


@fixture
def pod_name() -> str:
    """Pod name information."""
    return "pod_name"


@fixture
def pod_identifier(pod_name: str, pod_namespace: str) -> str:
    """Pod identifier information."""
    return f"{pod_namespace}/{pod_name}"


@unit_test
class TestBitfountHub:
    """Tests for the BitfountHub API."""

    # START OF FIXTURES #
    @fixture
    def good_pod(self, s3_download_url: _S3PresignedURL) -> _PodDetailsResponseJSON:
        """JSON response for a good get pod."""
        return {
            "podIdentifier": "123e4567-e89b-12d3-a456-426614174000",
            "podName": "blahhh",
            "podDisplayName": "None",
            "podPublicKey": "None",
            "accessManagerPublicKey": "None",
            "description": "None",
            "schemaStorageKey": "None",
            "isOnline": True,
            "providerUserName": "username",
            "visibility": "public",
            # This is appended if schemaStorageKey is not null. For our tests
            # we attach it regardless.
            "schemaDownloadUrl": s3_download_url,
        }

    @fixture
    def good_pods(self) -> List[_MultiPodDetailsResponseJSON]:
        """JSON response for get all pods.

        Matches the return value of getAllPods() in
        bitfount-web/bf-packages/bitfount-hub/modules/model/pods.js
        """
        return [
            {
                "podIdentifier": "123e4567-e89b-12d3-a456-426614174000",
                "name": "blahhh",
                "podDisplayName": "None",
                "isOnline": True,
                "podPublicKey": "None",
                "accessManagerPublicKey": "None",
                "description": "None",
                "providerUserName": "username",
                "podPagePath": "username/blahhh",
            }
        ]

    @fixture
    def model_missing_implementations(self) -> str:
        """Bitfount Model file with still abstract model."""
        return inspect.cleandoc(
            """
            from bitfount.models.bitfount_model import BitfountModel

            class MyModel(BitfountModel):
                pass
            """
        )

    @fixture
    def model_multiple_classes(self) -> str:
        """Bitfount Model file containing multiple classes."""
        return inspect.cleandoc(
            """
            from bitfount.models.bitfount_model import BitfountModel

            class MyModel(BitfountModel):
                def _set_metrics(self, metrics=None):
                    pass

                def serialize(self, filename):
                    pass

                def deserialize(self, filename):
                    pass

                def evaluate(self, test_dl=None):
                    pass

                def fit(self, *args, **kwargs):
                    pass

                def apply_weight_updates():
                    ...

                def backend_tensor_shim():
                    ...

                def diff_params():
                    ...

                def get_param_states():
                    ...

                def set_model_training_iterations():
                    ...

                def tensor_precision():
                    ...

                def update_params():
                    ...

                def _get_import_statements():
                    ...

                def _get_model():
                    ...

            class MyOtherModel(BitfountModel):
                def _set_metrics(self, metrics=None):
                    pass

                def serialize(self, filename):
                    pass

                def deserialize(self, filename):
                    pass

                def evaluate(self, test_dl=None):
                    pass

                def fit(self, *args, **kwargs):
                    pass

                def apply_weight_updates():
                    ...

                def backend_tensor_shim():
                    ...

                def diff_params():
                    ...

                def get_param_states():
                    ...

                def set_model_training_iterations():
                    ...

                def tensor_precision():
                    ...

                def update_params():
                    ...

                def _get_import_statements():
                    ...

                def _get_model():
                    ...
            """
        )

    @fixture
    def good_models_response(self) -> List[_MultiModelDetailsResponseJSON]:
        """Good model details in format for multiple models as fixture."""
        return [
            {
                "modellerName": "username",
                "modelName": "blah",
                "modelStorageKey": "model_storage_key",
                "modelHash": "nice_hash",
            },
        ]

    @fixture
    def good_model_response(
        self,
        bitfount_model_correct_structure: str,
        s3_download_url: _S3PresignedURL,
        tmp_path: Path,
    ) -> _ModelDetailsResponseJSON:
        """Good model code and correct hash as fixture."""
        model_file = tmp_path / "model.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        return {
            "modelStorageKey": "model_storage_key",
            "modelDownloadUrl": s3_download_url,
            "modelHash": hash_file_contents(model_file),
        }

    @fixture
    def json_success(self) -> _HubSuccessResponseJSON:
        """Hub success message in JSON format as fixture."""
        return {
            "success": True,
            "message": "No message",
        }

    @fixture
    def json_failure(self) -> _HubFailureResponseJSON:
        """Hub failure message in JSON format as fixture."""
        return {
            "success": False,
            "errorMessage": "Error",
        }

    @fixture
    def hub_url(self) -> str:
        """Test hub URL as fixture."""
        return "http://test.bitfount.com"

    @fixture
    def hub(self, hub_url: str, mocker: MockerFixture, tmp_path: Path) -> BitfountHub:
        """A BitfountHub instance as fixture."""
        session = BitfountSession()
        # Patch authenticated, username and user_storage_path onto session
        mocker.patch.object(
            type(session), "authenticated", PropertyMock(return_value=True)
        )
        mocker.patch.object(
            type(session), "username", PropertyMock(return_value="test")
        )
        mocker.patch.object(
            type(session), "access_token", PropertyMock(return_value="test")
        )
        session.user_storage_path = tmp_path

        # Treat this as a BitfountSession for these tests
        return BitfountHub(session, hub_url)

    @fixture
    def schema_json(self) -> _JSONDict:
        """Dictionary view of a false schema."""
        return {"I": "am", "a": "pod schema"}

    @fixture
    def schema_size(self, schema_json: _JSONDict) -> int:
        """Returns the object upload size of the schema."""
        return _get_packed_data_object_size(schema_json)

    @fixture
    def public_metadata(
        self, pod_name: str, schema_json: _JSONDict
    ) -> PodPublicMetadata:
        """Pod public metadata as fixture."""
        return PodPublicMetadata(
            pod_name,
            "pod_display_name",
            "pod_description",
            schema_json,
        )

    # END OF FIXTURES #

    def test_create_hub_without_bitfount_session(self, mocker: MockerFixture) -> None:
        """Tests that BitfountSession is created under the hood if not provided."""
        mock_session = Mock(
            spec=BitfountSession,
            authenticated=False,
            user_storage_path="",
            username="",
        )
        mocker.patch("bitfount.hub.api.BitfountSession", return_value=mock_session)
        hub = BitfountHub()
        mock_session.authenticate.assert_called_once()
        assert isinstance(hub.session, BitfountSession)

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod(
        self,
        good_pod: _PodDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: str,
        pod_name_: str,
        pod_namespace_: str,
    ) -> None:
        """Test Hub.get_pod() can retrieve pod details."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            json=good_pod,
        )

        pod_details = hub.get_pod(pod_identifier_, pod_namespace_, pod_name_)

        assert pod_details == good_pod

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_request_exception(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: str,
        pod_name_: str,
        pod_namespace_: str,
    ) -> None:
        """Test Hub.get_pod() returns None if RequestException is encountered."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body=RequestException("TEST"),
        )

        pod_details = hub.get_pod(pod_identifier_, pod_namespace_, pod_name_)

        # Pod details should be None due to exception
        assert pod_details is None
        # Check call was made
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @pytest.mark.parametrize("status_code", (400, 404, 500))
    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_bad_response(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Test Hub.get_pod() returns None if a non-OK response is returned."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            status=status_code,
        )

        pod_details = hub.get_pod(pod_identifier_, pod_namespace_, pod_name_)

        # Pod details should be None due to exception
        assert pod_details is None
        # Check call/retries were made
        expected_url_calls = 1
        if status_code in web_utils._RETRY_STATUS_CODES:
            expected_url_calls += web_utils._DEFAULT_MAX_RETRIES
        assert responses.assert_call_count(
            f"{hub_url}/api/pods/{pod_identifier}", expected_url_calls
        )
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert f"Bitfount Hub connection failed with: {status_code}" in error_logs

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_bad_json(
        self,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Test Hub.get_pod() raises an exception if no JSON in response."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body="NOT_JSON",
        )

        with pytest.raises(InvalidJSONError, match="Invalid JSON response "):
            hub.get_pod(pod_identifier_, pod_namespace_, pod_name_)

        # Check call was made
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_key(
        self,
        good_pod: _PodDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Checks get pod key works correctly."""
        # Change pod key to other value
        good_pod["podPublicKey"] = "KEY"

        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            json=good_pod,
        )

        key = hub.get_pod_key(pod_identifier_, pod_namespace_, pod_name_)

        assert key == "KEY"

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_key_request_exception(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Checks get pod key handles RequestException during request."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body=RequestException("TEST"),
        )

        key = hub.get_pod_key(pod_identifier_, pod_namespace_, pod_name_)

        # Returned value is empty if exception occurred
        assert key is None
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_key_bad_json(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Checks get pod key raises HTTPError when passed bad JSON."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body="NOT_JSON",
        )

        with pytest.raises(InvalidJSONError, match="Invalid JSON response "):
            hub.get_pod(pod_identifier_, pod_namespace_, pod_name_)

        # Check call was made
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)

    @pytest.mark.parametrize(
        argnames=["pod_identifier_", "pod_name_", "pod_namespace_"],
        argvalues=[
            [lazy_fixture("pod_identifier"), None, None],
            [None, lazy_fixture("pod_name"), lazy_fixture("pod_namespace")],
        ],
    )
    @responses.activate
    def test_bitfounthub_get_pod_key_no_key_in_json(
        self,
        caplog: LogCaptureFixture,
        good_pod: _PodDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
        # We use these names to avoid recursive parameterisation issues due to the
        # parametrize() call above.
        pod_identifier_: Optional[str],
        pod_name_: Optional[str],
        pod_namespace_: Optional[str],
    ) -> None:
        """Checks get pod key returns empty string if no key in JSON."""
        # Remove public key field from response
        good_pod.pop("podPublicKey")  # type: ignore[misc] # Reason: this is the test purpose # noqa: B950

        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            json=good_pod,
        )

        key = hub.get_pod_key(pod_identifier_, pod_namespace_, pod_name_)

        # Key should be None
        assert key is None
        # Check call was made
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Response JSON contained no public key: " in error_logs

    @responses.activate
    def test_get_all_pods_success(
        self,
        good_pods: List[_MultiPodDetailsResponseJSON],
        hub: BitfountHub,
        hub_url: str,
    ) -> None:
        """Checks get pods works correctly."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods",
            json=good_pods,
        )

        results = hub.get_all_pods()

        # Should only find a single pod
        assert results == good_pods
        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)

    @responses.activate
    def test_get_all_models_success(
        self,
        good_models_response: List[_MultiModelDetailsResponseJSON],
        hub: BitfountHub,
        hub_url: str,
    ) -> None:
        """Checks get models works correctly."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/models",
            json=good_models_response,
        )

        results = hub.get_all_models()

        # Should only return the single model
        assert results == good_models_response
        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)

    @responses.activate
    def test_get_all_pods_failure(
        self, caplog: LogCaptureFixture, hub: BitfountHub, hub_url: str
    ) -> None:
        """Checks empty list returned if get pods raises exception."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods",
            body=RequestException("TEST"),
        )

        results = hub.get_all_pods()

        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)
        # Returned value is empty if exception occurred
        assert len(results) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @responses.activate
    def test_get_all_models_failure(
        self, caplog: LogCaptureFixture, hub: BitfountHub, hub_url: str
    ) -> None:
        """Checks empty list returned if get models raises exception."""
        responses.add(
            responses.GET,
            f"{hub_url}/api/models",
            body=RequestException("TEST"),
        )

        results = hub.get_all_models()

        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)
        # Returned value is empty if exception occurred
        assert len(results) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @responses.activate
    def test_send_model_success(
        self,
        bitfount_model_correct_structure: str,
        bitfount_model_correct_structure_size: int,
        hub: BitfountHub,
        hub_url: str,
        mock_s3_file_upload_in_api_module: Mock,
        mocker: MockerFixture,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        tmp_path: Path,
    ) -> None:
        """Checks send model works correctly."""
        model_name = "MyModel"
        model_file = tmp_path / f"{model_name}.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        mock_hash = mocker.patch("bitfount.hub.api.hash_file_contents")
        file_hash = "file_hash"
        mock_hash.return_value = file_hash

        # Create fake hub response
        model_upload_response: _ModelUploadResponseJSON = {
            "uploadUrl": s3_upload_url,
            "uploadFields": s3_upload_fields,
            "success": True,
            "errorMessage": "None",
        }

        responses.add(
            responses.POST,
            f"{hub_url}/api/models",
            json=model_upload_response,
            match=[
                json_params_matcher(
                    {
                        "modelName": model_name,
                        "modelSize": bitfount_model_correct_structure_size,
                        "modelHash": file_hash,
                    }
                )
            ],
        )

        hub.send_model(model_file)

        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)
        # Check upload call was made to S3 correctly
        mock_s3_file_upload_in_api_module.assert_called_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            file_contents=bitfount_model_correct_structure.encode("utf-8"),
            file_name=model_file.name,
        )

    @responses.activate
    def test_send_model_fails_too_large(
        self, hub: BitfountHub, mocker: MockerFixture
    ) -> None:
        """Tests that send_model() fails if the model file is too large."""
        # Mock out model verification so it will pass
        mocker.patch.object(hub, "_verify_bitfount_model_format")
        mock_hash = mocker.patch("bitfount.hub.api.hash_file_contents")

        # Create mock file path and set the "content" to be slightly too big
        too_long_size: int = _MAX_CUSTOM_MODEL_SIZE_BYTES + 1
        mock_code_path = Mock()
        mock_code_path.read_text.return_value = "a" * too_long_size

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Model is too large to upload: expected max "
                f"{_MAX_CUSTOM_MODEL_SIZE_MEGABYTES} megabytes, got "
                f"{_get_mb_from_bytes(too_long_size).fractional} megabytes."
            ),
        ):
            hub.send_model(mock_code_path)

        # Check that call to hub didn't occur before error occurred
        assert not responses.calls
        mock_hash.assert_called_once_with(mock_code_path)

    @responses.activate
    def test_send_model_missing_implementations(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        model_missing_implementations: str,
        tmp_path: Path,
    ) -> None:
        """Checks send fails if model is still abstract."""
        model_file = tmp_path / "MyModel.py"
        model_file.touch()
        model_file.write_text(model_missing_implementations)

        with pytest.raises(ModelUploadError, match="Model incorrectly structured"):
            hub.send_model(model_file)

        # Check that no request was made (should have stopped before this point)
        assert len(responses.calls) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert (
            "Model incorrectly structured. Error: Subclass of `BitfountModel` "
            "not found in file or is still abstract." in error_logs
        )

    @responses.activate
    def test_send_model_incorrect_format(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        model_multiple_classes: str,
        tmp_path: Path,
    ) -> None:
        """Checks send fails if multiple classes."""
        model_file = tmp_path / "MyModel.py"
        model_file.touch()
        model_file.write_text(model_multiple_classes)

        with pytest.raises(ModelUploadError, match="Model incorrectly structured"):
            hub.send_model(model_file)

        # Check that no request was made (should have stopped before this point)
        assert len(responses.calls) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert (
            "Model incorrectly structured. Error: Model file contains 2 models. "
            "Must be just 1." in error_logs
        )

    @responses.activate
    def test_send_model_incorrect_name(
        self,
        bitfount_model_correct_structure: str,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        tmp_path: Path,
    ) -> None:
        """Checks send fails if class name and filename don't match."""
        model_file = tmp_path / "NotaMatchingName.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)

        with pytest.raises(ModelUploadError, match="Model incorrectly structured."):
            hub.send_model(model_file)

        # Check that no request was made (should have stopped before this point)
        assert len(responses.calls) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert (
            "Model incorrectly structured. Error: MyModel != NotaMatchingName. "
            "Model class name must be the same as the filename" in error_logs
        )

    @responses.activate
    def test_send_model_handles_import_error(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        mocker: MockerFixture,
    ) -> None:
        """Checks send fails if unable to import."""
        mocker.patch.object(
            hub,
            "_verify_bitfount_model_format",
            side_effect=ImportError("IMPORT ERROR"),
        )

        with pytest.raises(ModelUploadError, match="Unable to import model."):
            hub.send_model(model_code_path=Mock())

        # Check that no request was made (should have stopped before this point)
        assert len(responses.calls) == 0
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "IMPORT ERROR" in error_logs

    @responses.activate
    def test_send_model_failure(
        self,
        bitfount_model_correct_structure: str,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        json_failure: _HubFailureResponseJSON,
        mocker: MockerFixture,
        tmp_path: Path,
    ) -> None:
        """Checks send model handles failure to upload."""
        model_name = "MyModel"
        model_file = tmp_path / f"{model_name}.py"
        mock_hash = mocker.patch("bitfount.hub.api.hash_file_contents")
        file_hash = "file_hash"
        mock_hash.return_value = file_hash
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)

        responses.add(
            responses.POST,
            f"{hub_url}/api/models",
            json=json_failure,
            match=[
                json_params_matcher(
                    {
                        "modelName": model_name,
                        "modelSize": len(
                            bitfount_model_correct_structure.encode("utf-8")
                        ),
                        "modelHash": file_hash,
                    }
                )
            ],
        )

        with pytest.raises(
            ModelUploadError, match="Failed to upload model details to hub"
        ):
            hub.send_model(model_file)

        # Check that correct request was made (even though it failed))
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert (
            "Could not send model to Bitfount Hub. Failed with message: Error"
            in error_logs
        )

    @responses.activate
    def test_send_model_request_exception(
        self,
        bitfount_model_correct_structure: str,
        bitfount_model_correct_structure_size: int,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        mocker: MockerFixture,
        tmp_path: Path,
    ) -> None:
        """Checks request exception handled and logged if raised when sending."""
        model_name = "MyModel"
        model_file = tmp_path / f"{model_name}.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        mock_hash = mocker.patch("bitfount.hub.api.hash_file_contents")
        file_hash = "file_hash"
        mock_hash.return_value = file_hash

        responses.add(
            responses.POST,
            f"{hub_url}/api/models",
            body=RequestException("TEST"),
            match=[
                json_params_matcher(
                    {
                        "modelName": model_name,
                        "modelSize": bitfount_model_correct_structure_size,
                        "modelHash": file_hash,
                    }
                )
            ],
        )

        with pytest.raises(
            ModelUploadError,
            match="Request exception occurred when uploading model details to hub",
        ):
            hub.send_model(model_file)

        # Check that correct request was made (even though it failed))
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @pytest.mark.parametrize("s3_exception_cls", (RequestException, HTTPError))
    @responses.activate
    def test_send_model_handles_s3_exception(
        self,
        bitfount_model_correct_structure: str,
        bitfount_model_correct_structure_size: int,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        mock_s3_file_upload_in_api_module: Mock,
        mocker: MockerFixture,
        s3_exception_cls: Type[Exception],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        tmp_path: Path,
    ) -> None:
        """Checks send model handles an exception when uploading to S3."""
        model_name = "MyModel"
        model_file = tmp_path / f"{model_name}.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        mock_hash = mocker.patch("bitfount.hub.api.hash_file_contents")
        file_hash = "file_hash"
        mock_hash.return_value = file_hash

        # Create fake hub response
        model_upload_response: _ModelUploadResponseJSON = {
            "uploadUrl": s3_upload_url,
            "uploadFields": s3_upload_fields,
            "success": True,
            "errorMessage": "None",
        }

        responses.add(
            responses.POST,
            f"{hub_url}/api/models",
            json=model_upload_response,
            match=[
                json_params_matcher(
                    {
                        "modelName": model_name,
                        "modelSize": bitfount_model_correct_structure_size,
                        "modelHash": file_hash,
                    }
                )
            ],
        )

        # Add exception to S3 upload
        mock_s3_file_upload_in_api_module.side_effect = s3_exception_cls("S3 ERROR")

        with pytest.raises(ModelUploadError, match="Failed to upload model to S3"):
            hub.send_model(model_file)

        # Check that correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/models", 1)
        # Check upload call was made to S3 correctly
        mock_s3_file_upload_in_api_module.assert_called_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            file_contents=bitfount_model_correct_structure.encode("utf-8"),
            file_name=model_file.name,
        )
        # Check logs
        error_logs = get_error_logs(caplog)
        assert "Failed to upload model to S3: S3 ERROR" in error_logs

    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_success(
        self,
        bitfount_model_correct_structure: str,
        good_model_response: _ModelDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        mock_s3_file_download_in_api_module: Mock,
        model_version: Optional[int],
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Checks get model works correctly."""
        # Mock S3 download return value
        mock_s3_file_download_in_api_module.return_value = (
            bitfount_model_correct_structure
        )
        url = f"{hub_url}/api/models?modellerName=username&modelName=MyModel"
        if model_version:
            good_model_response.update({"modelVersion": model_version})
            url = url + f"&modelVersion={model_version}"

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json=good_model_response,
        )
        response = hub.get_model("username", "MyModel", model_version)

        # Returned model is of correct type
        assert response is not None
        assert issubclass(response, BitfountModel)
        # Check correct request was made
        assert responses.assert_call_count(url, 1)
        # Check correct S3 download was made
        mock_s3_file_download_in_api_module.assert_called_once_with(
            s3_download_url, encoding="utf-8"
        )

    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_none_returned(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        model_version: Optional[int],
    ) -> None:
        """Checks None returned if no model was found."""
        modeller_name = "username"
        model_name = "blah"
        url = (
            f"{hub_url}/api/models?modellerName={modeller_name}&modelName={model_name}"
        )
        if model_version:
            url = url + f"&modelVersion={model_version}"

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json=[],
        )

        response = hub.get_model("username", "blah", model_version)

        # Check that no model was returned
        assert response is None
        # Check correct request was made
        assert responses.assert_call_count(
            url,
            1,
        )
        # Assert warning message
        warning_logs = get_warning_logs(caplog)
        assert (
            f"No models registered by the name of {model_name} "
            f"from user {modeller_name}" in warning_logs
        )

    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_wrong_json_dict(
        self,
        hub: BitfountHub,
        hub_url: str,
        json_failure: _HubFailureResponseJSON,
        model_version: Optional[int],
    ) -> None:
        """Checks error raised if wrong JSON dict returned.

        Expected inner exception is KeyError.
        """
        modeller_name = "username"
        model_name = "blah"
        url = (
            f"{hub_url}/api/models?modellerName={modeller_name}&modelName={model_name}"
        )
        if model_version:
            url = url + f"&modelVersion={model_version}"

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json=json_failure,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Cannot retrieve model, no model URL in pod response: {json_failure}"
            ),
        ):
            hub.get_model(modeller_name, model_name, model_version)

    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_wrong_json_type(
        self,
        hub: BitfountHub,
        hub_url: str,
        json_failure: _HubFailureResponseJSON,
        model_version: Optional[int],
    ) -> None:
        """Checks error raised if wrong JSON type returned.

        Expected inner exception is TypeError.
        """
        modeller_name = "username"
        model_name = "blah"
        url = (
            f"{hub_url}/api/models?modellerName={modeller_name}&modelName={model_name}"
        )
        if model_version:
            url = url + f"&modelVersion={model_version}"

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json=[json_failure],  # list rather than dict
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Cannot retrieve model, no model URL in pod response: {[json_failure]}"
            ),
        ):
            hub.get_model(modeller_name, model_name, model_version)

    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_failure(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        model_version: Optional[int],
    ) -> None:
        """Checks get model handles RequestException being raised."""
        url = f"{hub_url}/api/models?modellerName=username&modelName=MyModel"
        if model_version:
            url += f"&modelVersion={model_version}"

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            body=RequestException("TEST"),
        )

        response = hub.get_model("username", "MyModel", model_version)

        # Check that no model was returned
        assert response is None
        # Check correct request was made
        assert responses.assert_call_count(url, 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @pytest.mark.skip("Reintroduce after [BIT-1073] is implemented.")
    @pytest.mark.parametrize("model_version", [None, 1])
    @responses.activate
    def test_get_model_hash_mismatch_raises_value_error(
        self,
        good_model_response: _ModelDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        model_version: Optional[int],
    ) -> None:
        """Checks exception raised when model code hash doesn't match."""
        # add a newline to the end of the model code to change hash
        good_model_response["code"] = good_model_response["code"] + "\n"  # type: ignore[typeddict-item]  # Reason: TODO: [BIT-1073] Remove ignore and fix issues  # noqa: B950

        responses.add(
            responses.GET,
            f"{hub_url}/api/models?modellerName=username&modelName=MyModel",
            match_querystring=True,
            json=good_model_response,
        )

        with pytest.raises(ValueError):
            hub.get_model("username", "MyModel", model_version)

        # Check correct request was made
        assert responses.assert_call_count(
            f"{hub_url}/api/models?modellerName=username&modelName=MyModel", 1
        )

    @responses.activate
    def test_register_pod_success(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        hub_url: str,
        mock_s3_data_upload_in_api_module: Mock,
        pod_name: str,
        public_metadata: PodPublicMetadata,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        schema_size: int,
    ) -> None:
        """Checks pod registration works correctly."""
        pod_public_key_str: str = authoriser_public_key.public_bytes(
            Encoding.OpenSSH, PublicFormat.OpenSSH
        ).decode("utf-8")
        access_manager_public_key_str: str = access_manager_public_key.public_bytes(
            Encoding.OpenSSH, PublicFormat.OpenSSH
        ).decode("utf-8")

        responses.add(
            responses.POST,
            f"{hub_url}/api/pods",
            json={
                "success": True,
                "uploadUrl": s3_upload_url,
                "uploadFields": s3_upload_fields,
            },
            match=[
                json_params_matcher(
                    {
                        "name": pod_name,
                        "podPublicKey": pod_public_key_str,
                        "accessManagerPublicKey": access_manager_public_key_str,
                        "podDisplayName": public_metadata.display_name,
                        "description": public_metadata.description,
                        "schemaSize": schema_size,
                    }
                )
            ],
        )

        hub.register_pod(
            public_metadata, authoriser_public_key, access_manager_public_key
        )

        # Check correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)
        # Assert S3 upload was made for schema
        mock_s3_data_upload_in_api_module.assert_called_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            data=public_metadata.schema,
        )

    @responses.activate
    def test_register_pod_failed(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        hub_url: str,
        mock_s3_data_upload_in_api_module: Mock,
        public_metadata: PodPublicMetadata,
    ) -> None:
        """Checks pod registration handles "failed" response from hub."""
        responses.add(
            responses.POST,
            f"{hub_url}/api/pods",
            json={"success": False},
            status=400,
        )

        with pytest.raises(HTTPError, match="Bitfount Hub connection failed with: 400"):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

        # Check that we haven't attempted to upload the schema
        mock_s3_data_upload_in_api_module.assert_not_called()

    @responses.activate
    def test_register_pod_bad_json(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        hub_url: str,
        public_metadata: PodPublicMetadata,
    ) -> None:
        """Checks exception raised on `body`, not `json`, response."""
        non_json_response = "NOT JSON RESPONSE"
        responses.add(
            responses.POST,
            f"{hub_url}/api/pods",
            body=non_json_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(f"Invalid JSON response (200): {non_json_response}"),
        ):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

    def test_register_pod_fails_schema_too_large(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        mocker: MockerFixture,
        public_metadata: PodPublicMetadata,
    ) -> None:
        """Tests that exception raised if schema is too large for storage."""
        # Patch out schema size calculation so we can set it
        too_large_size = _MAX_SCHEMA_SIZE_BYTES + 1
        mock_data_sizer = mocker.patch("bitfount.hub.api._get_packed_data_object_size")
        mock_data_sizer.return_value = too_large_size

        with pytest.raises(
            SchemaUploadError,
            match=re.escape(
                "Schema is too large to upload: "
                f"expected max {_MAX_SCHEMA_SIZE_MEGABYTES} megabytes, "
                f"got {_get_mb_from_bytes(too_large_size).fractional} megabytes."
            ),
        ):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

    @responses.activate
    def test_do_pod_heartbeat_success(
        self,
        authoriser_public_key: RSAPublicKey,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_name: str,
    ) -> None:
        """Checks do_pod_heartbeat correctly."""
        caplog.set_level(logging.DEBUG)

        authoriser_public_key_str = authoriser_public_key.public_bytes(
            Encoding.OpenSSH, PublicFormat.OpenSSH
        ).decode("utf-8")

        responses.add(
            responses.PATCH,
            f"{hub_url}/api/pods",
            json={"success": True},
            match=[
                json_params_matcher(
                    {"name": pod_name, "podPublicKey": authoriser_public_key_str}
                )
            ],
        )

        hub.do_pod_heartbeat(pod_name, authoriser_public_key)

        # Check correct request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)
        # Assert debug message
        debug_logs = get_debug_logs(caplog)
        assert "Status ping successful" in debug_logs

    @responses.activate
    def test_do_pod_heartbeat_failed_bad_status_code(
        self,
        authoriser_public_key: RSAPublicKey,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_name: str,
    ) -> None:
        """Test do_pod_heartbeat raises exception on 400+ status code."""
        responses.add(
            responses.PATCH,
            f"{hub_url}/api/pods",
            json={"success": False},
            status=400,
        )

        with pytest.raises(HTTPError, match="400 Client Error: Bad Request for url"):
            hub.do_pod_heartbeat(pod_name, authoriser_public_key)

        # Check request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)

    @responses.activate
    def test_do_pod_heartbeat_failed_not_ok_status_code(
        self,
        authoriser_public_key: RSAPublicKey,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_name: str,
    ) -> None:
        """Test do_pod_heartbeat skips "failed" (non-200) response from hub."""
        responses.add(
            responses.PATCH,
            f"{hub_url}/api/pods",
            json={"success": False},
            status=201,
        )

        hub.do_pod_heartbeat(pod_name, authoriser_public_key)

        # Check request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)
        # Assert warning message
        warning_logs = get_warning_logs(caplog)
        assert "Status update failed with (201):" in warning_logs

    @responses.activate
    def test_do_pod_heartbeat_bad_json(
        self,
        authoriser_public_key: RSAPublicKey,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_name: str,
    ) -> None:
        """Checks exception raised on `body`, not `json`, response."""
        non_json_response = "NOT JSON RESPONSE"

        responses.add(
            responses.PATCH,
            f"{hub_url}/api/pods",
            body=non_json_response,
            status=200,
        )

        hub.do_pod_heartbeat(pod_name, authoriser_public_key)

        # Check request was made
        assert responses.assert_call_count(f"{hub_url}/api/pods", 1)
        # Assert warning message
        warning_logs = get_warning_logs(caplog)
        assert f"Invalid JSON response (200): {non_json_response}" in warning_logs

    @responses.activate
    def test_add_pod_handles_http_500_error_when_registering(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        hub_url: str,
        json_failure: _HubFailureResponseJSON,
        public_metadata: PodPublicMetadata,
        remove_web_retry_backoff_sleep: Optional[Mock],
    ) -> None:
        """Tests 500 response correctly handled during pod registration."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            responses.POST,
            f"{hub_url}/api/pods",
            json=json_failure,
            status=500,
        )

        with pytest.raises(
            HTTPError,
            match="Bitfount Hub connection failed with: 500 Server Error",
        ):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

        # Check retries occurred
        assert (
            remove_web_retry_backoff_sleep.call_count == web_utils._DEFAULT_MAX_RETRIES
        )

    @responses.activate
    def test_add_pod_handles_http_non_200_201_error_when_registering(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        hub_url: str,
        json_failure: _HubFailureResponseJSON,
        public_metadata: PodPublicMetadata,
    ) -> None:
        """Tests non-200/201 response correctly handled during pod registration."""
        responses.add(
            responses.POST,
            f"{hub_url}/api/pods",
            json=json_failure,
            status=203,
        )

        with pytest.raises(
            HTTPError,
            match=re.escape(f"Unexpected response (203): {json.dumps(json_failure)}"),
        ):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

    @responses.activate
    def test_worker_handles_connection_error_when_registering(
        self,
        access_manager_public_key: RSAPublicKey,
        authoriser_public_key: RSAPublicKey,
        hub: BitfountHub,
        public_metadata: PodPublicMetadata,
        remove_web_retry_backoff_sleep: Optional[Mock],
    ) -> None:
        """Checks connection error raised during failed registering."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        # No `response` added, so responses raises a connection error. This is
        # wrapped with the RequestException raised by Hub.
        with pytest.raises(
            requests.exceptions.ConnectionError,
            match="Bitfount Hub connection failed with: Connection refused by",
        ):
            hub.register_pod(
                public_metadata, authoriser_public_key, access_manager_public_key
            )

        # Check retries occurred
        assert (
            remove_web_retry_backoff_sleep.call_count == web_utils._DEFAULT_MAX_RETRIES
        )

    @responses.activate
    def test_get_pod_schema(
        self,
        good_pod: _PodDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        mock_bitfount_schema_load: Mock,
        mock_s3_data_download_in_api_module: Mock,
        pod_identifier: str,
        schema_json: _JSONDict,
    ) -> None:
        """Test get_pod_schema() works correctly."""
        # Mock out S3 downloading of schema
        mock_s3_data_download_in_api_module.return_value = schema_json

        # Mock request response of pod details
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            json=good_pod,
        )

        schema = hub.get_pod_schema(pod_identifier)

        # Check schema is as expected
        assert schema == schema_json

    @responses.activate
    def test_get_pod_schema_request_exception(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
    ) -> None:
        """Test get_pod_schema() raises exception when RequestException() thrown."""
        # Mock request response of pod details
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body=RequestException("TEST"),
        )

        with pytest.raises(
            InvalidJSONError,
            match="Cannot retrieve pod schema, no schema URL in pod response: None",
        ):
            hub.get_pod_schema(pod_identifier)

        # Check URL was called
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)
        # Assert error message
        error_logs = get_error_logs(caplog)
        assert "Bitfount Hub connection failed with: TEST" in error_logs

    @responses.activate
    def test_get_pod_schema_bad_json(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
    ) -> None:
        """Test get_pod_schema() raises exception when bad JSON received."""
        # Mock request response of pod details
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            body="NOT_JSON",
        )

        with pytest.raises(InvalidJSONError, match="Invalid JSON response "):
            hub.get_pod_schema(pod_identifier)

        # Check URL was called
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)

    @responses.activate
    def test_get_pod_schema_no_schema_url(
        self,
        caplog: LogCaptureFixture,
        good_pod: _PodDetailsResponseJSON,
        hub: BitfountHub,
        hub_url: str,
        pod_identifier: str,
    ) -> None:
        """Test get_pod_schema() raises exception when no download URL in response."""
        # Remove schemaDownloadUrl from pod response
        good_pod.pop("schemaDownloadUrl")  # type: ignore[misc] # Reason: purpose of test # noqa: B950

        # Mock request response of pod details
        responses.add(
            responses.GET,
            f"{hub_url}/api/pods/{pod_identifier}",
            json=good_pod,
        )

        with pytest.raises(
            InvalidJSONError,
            match=f"Cannot retrieve pod schema, no schema URL in pod response: "
            f"{good_pod}",
        ):
            hub.get_pod_schema(pod_identifier)

        # Check URL was called
        assert responses.assert_call_count(f"{hub_url}/api/pods/{pod_identifier}", 1)

    @fixture
    def monitor_update(self) -> _MonitorPostJSON:
        """Example monitor update in JSON format."""
        return {
            "taskId": "task-id",
            "senderId": "sender-mailbox-id",
            "recipientId": "recipient-mailbox-id",
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "privacy": "PRIVATE",
            "type": "TASK_STATUS_UPDATE",
            "message": "this is a message",
            "metadata": {"meta": "data"},
            "progress": {"progress": {"value": 1, "total": 10}},
            "resourceUsage": {"a_resource": 10},
        }

    @fixture
    def ingestion_endpoint_url(self, hub_url: str) -> str:
        """Monitor service ingestion endpoint url."""
        return f"{hub_url}/api/ingest"

    @responses.activate
    def test_send_monitor_update(
        self,
        hub: BitfountHub,
        ingestion_endpoint_url: str,
        monitor_update: _MonitorPostJSON,
    ) -> None:
        """Test send_monitor_update works correctly."""
        responses.add(
            responses.POST,
            url=ingestion_endpoint_url,
            status=200,
            match=[json_params_matcher(cast(_JSONDict, monitor_update))],
        )

        hub.send_monitor_update(monitor_update)

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=(400, 404, 500),
    )
    @responses.activate
    def test_send_monitor_update_http_error(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        ingestion_endpoint_url: str,
        monitor_update: _MonitorPostJSON,
        status_code: int,
    ) -> None:
        """Test send_monitor_update handles HTTP errors."""
        responses.add(
            responses.POST,
            url=ingestion_endpoint_url,
            status=status_code,
            match=[json_params_matcher(cast(_JSONDict, monitor_update))],
        )

        with pytest.raises(HTTPError):
            hub.send_monitor_update(monitor_update)

        # Check error logs
        error_logs = get_error_logs(caplog)
        assert str(status_code) in error_logs

    @responses.activate
    def test_send_monitor_update_request_exception(
        self,
        caplog: LogCaptureFixture,
        hub: BitfountHub,
        ingestion_endpoint_url: str,
        monitor_update: _MonitorPostJSON,
    ) -> None:
        """Test send_monitor_update handles request errors."""
        responses.add(
            responses.POST,
            url=ingestion_endpoint_url,
            body=RequestException("test error"),
            match=[json_params_matcher(cast(_JSONDict, monitor_update))],
        )

        with pytest.raises(RequestException, match="test error"):
            hub.send_monitor_update(monitor_update)

        # Check error logs
        error_logs = get_error_logs(caplog)
        assert "test error" in error_logs


class TestAccessManager:
    """Test the BitfountAM class."""

    @fixture
    def url(self) -> str:
        """Access Manager URL as a fixture."""
        return "http://test-am.hub.bitfount.com"

    @fixture
    def session(self, mocker: MockerFixture) -> BitfountSession:
        """A session with auth."""
        session = BitfountSession()
        # Patching `authenticated` status onto session as it is required
        mocker.patch.object(
            type(session), "authenticated", PropertyMock(return_value=True)
        )
        mocker.patch.object(
            type(session), "access_token", PropertyMock(return_value="test")
        )
        session._access_token = "test"
        session.headers.update({"x-Bitfount-unit-test": "Session was used"})
        return session

    @fixture
    def access_manager(self, session: BitfountSession, url: str) -> BitfountAM:
        """Access Manager instance as a fixture."""
        return BitfountAM(session, url)

    @fixture
    def expected_saml_challenge(self) -> str:
        """SAML Challenge string from access manager."""
        return "a legit saml challenge"

    @fixture
    def expected_saml_request_id(self) -> str:
        """SAML Request ID from access manager."""
        return "my first saml"

    @fixture
    def saml_endpoint_result(
        self, expected_saml_challenge: str, expected_saml_request_id: str
    ) -> _SAMLChallengeResponseJSON:
        """Result of GET request to AM SAML endpoint."""
        return {"id": expected_saml_request_id, "samlRequest": expected_saml_challenge}

    @fixture
    def modeller_name(self) -> str:
        """Modeller name."""
        return "someModeller"

    @fixture
    def modeller_protocol_request(self) -> SerializedProtocol:
        """Modeller request."""
        return {
            "class_name": "bitfount.FederatedAveraging",
            "algorithm": {"class_name": "bitfount.FederatedModelTraining"},
        }

    @fixture
    def fake_original_saml_request_id(self) -> str:
        """Original SAML Request ID for tests."""
        return "12345"

    @fixture
    def expected_saml_post_body(
        self,
        fake_original_saml_request_id: str,
        modeller_name: str,
        modeller_protocol_request: SerializedProtocol,
        pod_identifier: str,
    ) -> _SAMLAdditionalInfoPOSTJSON:
        """JSON body Pod sends to AM to check access."""
        saml_post_body: _SAMLAdditionalInfoPOSTJSON = {
            "originalSAMLRequestID": fake_original_saml_request_id,
            "podIdentifier": pod_identifier,
            "modellerName": modeller_name,
            "modellerProtocolRequest": modeller_protocol_request,
            "identityProvider": "SAML",
        }
        return saml_post_body

    @fixture
    def fake_access_token(self) -> str:
        """Fake access token."""
        return "fake_access_token"

    @fixture
    def expected_oidc_post_body(
        self,
        fake_access_token: str,
        modeller_name: str,
        modeller_protocol_request: SerializedProtocol,
        pod_identifier: str,
    ) -> _OIDCAccessCheckPostJSON:
        """JSON body for OIDC Access Check POST request."""
        return {
            "podIdentifier": pod_identifier,
            "modellerName": modeller_name,
            "modellerProtocolRequest": modeller_protocol_request,
            "modellerToken": fake_access_token,
            "identityProvider": "OIDC",
        }

    @fixture
    def unsigned_task(self) -> bytes:
        """Unsigned task."""
        return b"unsigned_task"

    @fixture
    def signed_task(self) -> bytes:
        """Signed task."""
        return b"signed_task"

    @fixture
    def expected_signature_based_access_post_body(
        self,
        modeller_name: str,
        modeller_protocol_request: SerializedProtocol,
        pod_identifier: str,
        signed_task: bytes,
        unsigned_task: bytes,
    ) -> _SignatureBasedAccessCheckPostJSON:
        """JSON body for Signature based Access Check POST request."""
        return {
            "podIdentifier": pod_identifier,
            "modellerName": modeller_name,
            "modellerRequest": modeller_protocol_request,
            "unsignedTask": unsigned_task,
            "signedTask": signed_task,
            "identityProvider": "SIGNATURE",
        }

    @unit_test
    def test_create_access_manager_without_bitfount_session(
        self, mocker: MockerFixture
    ) -> None:
        """Tests that BitfountSession is created under the hood if not provided."""
        mock_session = Mock(
            spec=BitfountSession,
            authenticated=False,
            user_storage_path="",
            username="",
        )
        mocker.patch("bitfount.hub.api.BitfountSession", return_value=mock_session)
        hub = BitfountAM()
        mock_session.authenticate.assert_called_once()
        assert isinstance(hub.session, BitfountSession)

    @unit_test
    @responses.activate
    def test_get_access_manager_key(
        self,
        access_manager: BitfountAM,
        access_manager_public_key: RSAPublicKey,
        url: str,
    ) -> None:
        """Checks AM key getter works."""
        expected_response: _AccessManagerKeyResponseJSON = {
            "accessManagerPublicKey": access_manager_public_key.public_bytes(
                Encoding.OpenSSH, PublicFormat.OpenSSH
            ).decode("utf-8"),
        }

        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            json=expected_response,
            status=200,
        )

        key = access_manager.get_access_manager_key()

        # Check key type and value are what we expect
        assert isinstance(key, RSAPublicKey)
        assert _RSAEncryption.serialize_public_key(
            key
        ) == _RSAEncryption.serialize_public_key(access_manager_public_key)

    @integration_test
    @pytest.mark.timeout(10)
    def test_get_access_manager_key_speed(self, monkeypatch: MonkeyPatch) -> None:
        """Checks AM key getter is able to retrieve the key with minimal delay.

        This test is to ensure we don't regress due to IPv4/IPv6 issues either on the
        requests side or the AM side.
        """
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")
        mock_session = Mock(authenticated=True, spec=BitfountSession)
        am = BitfountAM(session=mock_session)
        key = am.get_access_manager_key()
        # Check key type and value are what we expect
        assert isinstance(key, RSAPublicKey)
        # Assert that the authenticate method on the session was not called
        mock_session.authenticate.assert_not_called()

    @unit_test
    @responses.activate
    def test_get_access_manager_key_fails_request_exception(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised on if request raises RequestException."""
        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            body=RequestException("TEST"),
        )

        with pytest.raises(
            RequestException,
            match="Bitfount Access Manager connection failed with: TEST.$",
        ):
            access_manager.get_access_manager_key()

    @unit_test
    @pytest.mark.parametrize("status_code", (400, 404, 500))
    @responses.activate
    def test_get_access_manager_key_fails_non_200(
        self,
        access_manager: BitfountAM,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        url: str,
    ) -> None:
        """Checks exception raised on non-200 response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            status=status_code,
        )

        with pytest.raises(HTTPError):
            access_manager.get_access_manager_key()

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @unit_test
    @responses.activate
    def test_get_access_manager_key_fails_non_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised if response is `body`, not `json`."""
        non_json_response = "NOT JSON RESPONSE"
        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            body=non_json_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(f"Invalid JSON response (200): {non_json_response}"),
        ):
            access_manager.get_access_manager_key()

    @unit_test
    @responses.activate
    def test_get_access_manager_key_fails_no_key_in_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised if key not in JSON."""
        expected_response = {
            "incorrectKey": "incorrectValue",
        }

        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            json=expected_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Unable to extract public key from access manager response, "
                f"no key in JSON: {expected_response}"
            ),
        ):
            access_manager.get_access_manager_key()

    @unit_test
    @responses.activate
    def test_get_access_manager_key_fails_bad_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks bad JSON response causes exception."""
        expected_response = "JSON BUT AT WHAT COST"

        responses.add(
            responses.GET,
            f"{url}/api/access-manager-key",
            json=expected_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Unable to extract public key from access manager response, "
                f"no key in JSON: {expected_response}"
            ),
        ):
            access_manager.get_access_manager_key()

    @unit_test
    @responses.activate
    def test_get_saml_challenge(
        self,
        access_manager: BitfountAM,
        expected_saml_challenge: str,
        expected_saml_request_id: str,
        saml_endpoint_result: _SAMLChallengeResponseJSON,
        session: BitfountSession,
        url: str,
    ) -> None:
        """Checks SAML challenge retrieval works."""
        responses.add(
            responses.GET,
            f"{url}/api/saml?idp=cli",
            json=saml_endpoint_result,
            status=200,
        )

        saml_challenge, saml_request_id = access_manager.get_saml_challenge()

        assert saml_challenge == expected_saml_challenge
        assert saml_request_id == expected_saml_request_id
        # Make sure the access token is added as a header to the request
        assert (
            responses.calls[0].request.headers["Authorization"]
            == f"Bearer {session.access_token}"
        )

    @unit_test
    @responses.activate
    def test_get_saml_challenge_fails_request_exception(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised on if request raises RequestException."""
        responses.add(
            responses.GET, f"{url}/api/saml?idp=cli", body=RequestException("TEST")
        )

        with pytest.raises(
            RequestException,
            match="Bitfount Access Manager connection failed with: TEST.$",
        ):
            access_manager.get_saml_challenge()

    @unit_test
    @pytest.mark.parametrize("status_code", (400, 404, 500))
    @responses.activate
    def test_get_saml_challenge_fails_non_200(
        self,
        access_manager: BitfountAM,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        url: str,
    ) -> None:
        """Checks exception raised on non-200 response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            responses.GET,
            f"{url}/api/saml?idp=cli",
            status=status_code,
            json={"error": "Some error"},
        )

        with pytest.raises(HTTPError):
            access_manager.get_saml_challenge()

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @unit_test
    @responses.activate
    def test_get_saml_challenge_fails_non_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised if response is `body`, not `json`."""
        non_json_response = "NOT JSON RESPONSE"
        responses.add(
            responses.GET,
            f"{url}/api/saml?idp=cli",
            body=non_json_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(f"Invalid JSON response (200): {non_json_response}"),
        ):
            access_manager.get_saml_challenge()

    @unit_test
    @responses.activate
    def test_get_saml_challenge_fails_wrong_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks exception raised if expected keys not in JSON."""
        expected_response = {
            "incorrectKey": "incorrectValue",
        }

        responses.add(
            responses.GET,
            f"{url}/api/saml?idp=cli",
            json=expected_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Unable to extract SAML Challenge from access manager response, "
                f"no challenge in JSON: {expected_response}"
            ),
        ):
            access_manager.get_saml_challenge()

    @unit_test
    @responses.activate
    def test_get_saml_challenge_fails_bad_json(
        self, access_manager: BitfountAM, url: str
    ) -> None:
        """Checks bad JSON response causes exception."""
        expected_response = "JSON BUT AT WHAT COST"

        responses.add(
            responses.GET,
            f"{url}/api/saml?idp=cli",
            json=expected_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(
                f"Unable to extract SAML Challenge from access manager response, "
                f"no challenge in JSON: {expected_response}"
            ),
        ):
            access_manager.get_saml_challenge()

    @unit_test
    @responses.activate
    def test_validate_saml_response_success(
        self,
        access_manager: BitfountAM,
        expected_saml_post_body: _SAMLAdditionalInfoPOSTJSON,
        session: BitfountSession,
        url: str,
    ) -> None:
        """Checks accepted SAML & Access check is returned."""
        accept_response: _AMAccessCheckResponseJSON = {"code": "ACCEPT"}
        saml_response: _SAMLResponse = {"saml": "response"}
        responses.add(
            responses.POST,
            f"{url}/api/access?idp=cli",
            json=accept_response,
        )

        assert (
            access_manager.validate_saml_response(
                saml_response,
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )
            == _PodResponseType.ACCEPT
        )

        request_body = responses.calls[0].request.body
        assert request_body is not None
        assert json.loads(request_body) == {**saml_response, **expected_saml_post_body}
        # Make sure we used the session to make the request!
        # Some headers are added when we attach the body
        # So we just check that the expected headers are a subset of the attached ones
        assert session.headers.items() <= responses.calls[0].request.headers.items()

    @unit_test
    @responses.activate
    def test_validate_saml_response_rejects_access(
        self, access_manager: BitfountAM, expected_saml_post_body: _JSONDict, url: str
    ) -> None:
        """Check validate SAML response when AM denies access."""
        reject_response: _AMAccessCheckResponseJSON = {"code": "UNAUTHORISED"}
        saml_response: _SAMLResponse = {"saml": "response"}
        responses.add(
            responses.POST,
            f"{url}/api/access?idp=cli",
            json=reject_response,
        )

        assert (
            access_manager.validate_saml_response(
                saml_response,
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )
            == _PodResponseType.UNAUTHORISED
        )

        request_body = responses.calls[0].request.body
        assert request_body is not None
        assert json.loads(request_body) == {**saml_response, **expected_saml_post_body}

    @unit_test
    @responses.activate
    def test_validate_saml_response_fails_request_exception(
        self, access_manager: BitfountAM, expected_saml_post_body: _JSONDict, url: str
    ) -> None:
        """Checks exception raised on request raising RequestException."""
        responses.add(
            responses.POST, f"{url}/api/access?idp=cli", body=RequestException("TEST")
        )

        with pytest.raises(
            RequestException,
            match="Bitfount Access Manager connection failed with: TEST.$",
        ):
            access_manager.validate_saml_response(
                {"saml": "response"},
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )

    @unit_test
    @pytest.mark.parametrize("status_code", (400, 404, 500))
    @responses.activate
    def test_validate_saml_response_fails_non_200(
        self,
        access_manager: BitfountAM,
        expected_saml_post_body: _JSONDict,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        url: str,
    ) -> None:
        """Checks exception raised on non-200 response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            responses.POST,
            f"{url}/api/access?idp=cli",
            status=status_code,
            json={"error": "Some error"},
        )

        with pytest.raises(HTTPError):
            access_manager.validate_saml_response(
                {"saml": "response"},
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @unit_test
    @responses.activate
    def test_validate_saml_response_fails_json_missing_response_code(
        self, access_manager: BitfountAM, expected_saml_post_body: _JSONDict, url: str
    ) -> None:
        """Checks "ERROR" code returned if expected keys not in JSON."""
        expected_response = {
            "incorrectKey": "incorrectValue",
        }

        responses.add(
            responses.POST,
            f"{url}/api/access?idp=cli",
            json=expected_response,
            status=200,
        )

        assert (
            access_manager.validate_saml_response(
                {"saml": "response"},
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )
            == _PodResponseType.NO_ACCESS
        )

    @unit_test
    @responses.activate
    def test_validate_saml_response_fails_wrong_json(
        self, access_manager: BitfountAM, expected_saml_post_body: _JSONDict, url: str
    ) -> None:
        """Checks exception raised if expected keys not in JSON."""
        expected_response = "JSON BUT AT WHAT COST"

        responses.add(
            responses.POST,
            f"{url}/api/access?idp=cli",
            json=expected_response,
            status=200,
        )

        with pytest.raises(
            InvalidJSONError,
            match=re.escape(f'Invalid JSON response (200): "{expected_response}"'),
        ):
            access_manager.validate_saml_response(
                {"saml": "response"},
                expected_saml_post_body["originalSAMLRequestID"],
                expected_saml_post_body["podIdentifier"],
                expected_saml_post_body["modellerName"],
                expected_saml_post_body["modellerProtocolRequest"],
            )

    @unit_test
    @responses.activate
    def test_check_oidc_access_request_success(
        self,
        access_manager: BitfountAM,
        expected_oidc_post_body: _OIDCAccessCheckPostJSON,
        fake_access_token: str,
        modeller_name: str,
        modeller_protocol_request: SerializedProtocol,
        pod_identifier: str,
        session: BitfountSession,
        url: str,
    ) -> None:
        """Checks OIDC access check works."""
        accept_response: _AMAccessCheckResponseJSON = {"code": "ACCEPT"}

        responses.add(
            responses.POST,
            f"{url}/api/access",
            json=accept_response,
        )

        assert (
            access_manager.check_oidc_access_request(
                pod_identifier=pod_identifier,
                serialized_protocol=modeller_protocol_request,
                modeller_name=modeller_name,
                modeller_access_token=fake_access_token,
            )
            == _PodResponseType.ACCEPT
        )

        request_body = responses.calls[0].request.body
        assert request_body is not None
        assert json.loads(request_body) == expected_oidc_post_body
        # Make sure we used the session to make the request!
        # Some headers are added when we attach the body
        # So we just check that the expected headers are a subset of the attached ones
        assert session.headers.items() <= responses.calls[0].request.headers.items()

    @unit_test
    @responses.activate
    def test_check_signature_based_access_request_success(
        self,
        access_manager: BitfountAM,
        expected_signature_based_access_post_body: _SignatureBasedAccessCheckPostJSON,
        modeller_name: str,
        modeller_protocol_request: SerializedProtocol,
        pod_identifier: str,
        session: BitfountSession,
        signed_task: bytes,
        unsigned_task: bytes,
        url: str,
    ) -> None:
        """Checks signatured based access check works."""
        accept_response: _AMAccessCheckResponseJSON = {"code": "ACCEPT"}

        responses.add(
            responses.POST,
            f"{url}/api/access",
            json=accept_response,
        )

        assert (
            access_manager.check_signature_based_access_request(
                unsigned_task=unsigned_task,
                signed_task=signed_task,
                pod_identifier=pod_identifier,
                serialized_protocol=modeller_protocol_request,
                modeller_name=modeller_name,
            )
            == _PodResponseType.ACCEPT
        )

        request_body = responses.calls[0].request.body
        assert request_body is not None

        # assert json.loads(request_body) == expected_signature_based_access_post_body
        # Make sure we used the session to make the request!
        # Some headers are added when we attach the body
        # So we just check that the expected headers are a subset of the attached ones
        assert session.headers.items() <= responses.calls[0].request.headers.items()

    @unit_test
    def test_am_url_gets_prod_url(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that the am is loaded based on production environment."""
        mock_session = Mock(
            spec=BitfountSession,
            authenticated=False,
            user_storage_path="",
            username="",
        )
        mocker.patch("bitfount.hub.api.BitfountSession", return_value=mock_session)
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "production")
        am = BitfountAM()
        assert am.access_manager_url == PRODUCTION_AM_URL

    @unit_test
    def test_am_url_gets_staging_url(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that the am url is loaded based on staging environment."""
        mock_session = Mock(
            spec=BitfountSession,
            authenticated=False,
            user_storage_path="",
            username="",
        )
        mocker.patch("bitfount.hub.api.BitfountSession", return_value=mock_session)
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")
        am = BitfountAM()
        assert am.access_manager_url == _STAGING_AM_URL

    @unit_test
    def test_am_url_gets_dev_url(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that the idp_url is loaded based on dev environment."""
        mock_session = Mock(
            spec=BitfountSession,
            authenticated=False,
            user_storage_path="",
            username="",
        )
        mocker.patch("bitfount.hub.api.BitfountSession", return_value=mock_session)
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "dev")
        am = BitfountAM()
        assert am.access_manager_url == _DEV_AM_URL


@unit_test
class TestCheckPodDetails:
    """Tests __check_pod_id_details function."""

    def test__check_pod_id_details_pod_identifier(self, pod_identifier: str) -> None:
        """Tests _check_pod_id_details works when given only pod_identifier."""
        returned_pod_id = _check_pod_id_details(pod_identifier=pod_identifier)
        assert returned_pod_id == pod_identifier

    def test__check_pod_id_details_pod_namespace_and_name(
        self, pod_identifier: str, pod_name: str, pod_namespace: str
    ) -> None:
        """Tests _check_pod_id_details works when given pod_namespace and pod_name.

        Check that DeprecationWarning is raised.
        """
        with pytest.warns(
            DeprecationWarning,
            match="pod_identifier should be used instead of pod_namespace and pod_name",
        ):
            returned_pod_id = _check_pod_id_details(
                pod_namespace=pod_namespace, pod_name=pod_name
            )
        assert returned_pod_id == pod_identifier

    def test__check_pod_id_details_none(self) -> None:
        """Tests _check_pod_id_details errors when given no args."""
        with pytest.raises(
            ValueError,
            match="At least one of pod_identifier OR pod_namespace and pod_name "
            "must be provided",
        ):
            _check_pod_id_details()

    def test__check_pod_id_details_one_of_namespace_or_name(
        self, pod_name: str, pod_namespace: str
    ) -> None:
        """Test _check_pod_id_details errors with only one of namespace or name."""
        # Only name
        with pytest.raises(
            ValueError,
            match="Both pod_namespace and pod_name must be provided, "
            "or neither must be.",
        ):
            _check_pod_id_details(pod_name=pod_name)

        # Only namespace
        with pytest.raises(
            ValueError,
            match="Both pod_namespace and pod_name must be provided, "
            "or neither must be.",
        ):
            _check_pod_id_details(pod_namespace=pod_namespace)

    def test__check_pod_id_details_all_provided(
        self, pod_identifier: str, pod_name: str, pod_namespace: str
    ) -> None:
        """Tests _check_pod_id_details works when all are provided and match.

        i.e. pod_identifier == pod_namespace/pod_name
        """
        returned_pod_id = _check_pod_id_details(pod_identifier, pod_namespace, pod_name)
        assert returned_pod_id == pod_identifier

    def test__check_pod_id_details_all_provided_but_not_match(
        self, pod_name: str, pod_namespace: str
    ) -> None:
        """Tests _check_pod_id_details errors when all are provided but don't match.

        i.e. pod_identifier != pod_namespace/pod_name
        """
        mismatch_pod_identifier = "not/theSame"

        with pytest.raises(
            ValueError,
            match="pod_identifier, pod_namespace and pod_name all provided, but "
            "pod_identifier doesn't match pod_namespace/pod_name.",
        ):
            _check_pod_id_details(mismatch_pod_identifier, pod_namespace, pod_name)



"""Classes concerning databunches."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

from bitfount.data.datafactory import _DataFactory, _get_default_data_factory
from bitfount.data.types import DataSplit, SemanticType
from bitfount.utils import _add_this_to_list

if TYPE_CHECKING:
    from bitfount.data.dataloaders import BitfountDataLoader
    from bitfount.data.datasets import _BaseBitfountDataset
    from bitfount.data.datasources.base_source import BaseSource
    from bitfount.data.datastructure import DataStructure
    from bitfount.data.schema import TableSchema

logger = logging.getLogger(__name__)


class BitfountDataBunch:
    """Wrapper for train, validation and test dataloaders.

    Provides methods to access dataloaders for training and evaluation. This is strictly
    a model concept and is not necessary for algorithms that do not have models.

    Args:
        data_structure: A `DataStructure` object.
        schema: A `TableSchema` object.
        datasource: A `BaseSource` object.
        data_factory: A `_DataFactory` instance for creating datasets and dataloaders.
            Defaults to None.
    """

    def __init__(
        self,
        data_structure: DataStructure,
        schema: TableSchema,
        datasource: BaseSource,
        data_factory: Optional[_DataFactory] = None,
    ):
        self.data_structure = data_structure
        self.schema = schema
        self.datasource = datasource
        self.data_factory = (
            data_factory if data_factory is not None else _get_default_data_factory()
        )

        # Column attributes
        self.target = self.data_structure.target
        self.loss_weights_col = self.data_structure.loss_weights_col
        self.multihead_col = self.data_structure.multihead_col
        self.ignore_classes_col = self.data_structure.ignore_classes_col

        # Placeholders for generated datasets
        self.train_ds: _BaseBitfountDataset  # training data is not optional
        self.validation_ds: Optional[_BaseBitfountDataset] = None
        self.test_ds: Optional[_BaseBitfountDataset] = None

        self.data_structure.set_training_column_split_by_semantic_type(self.schema)
        self._disallow_text_features()

        # TODO: [BIT-1167] This probably needs to be removed once we have implemented
        # dataset transformations. Currently, this call does nothing.
        self.datasource = self.data_structure.apply_dataset_transformations(
            self.datasource
        )
        self._load_data()
        self._create_datasets()

    def _load_data(self) -> None:
        """Loads the data from the datasource and applies dataset transformations."""
        kwargs = {}
        if isinstance(self.data_structure.query, str):
            kwargs["sql_query"] = self.data_structure.query

        if isinstance(self.data_structure.table, str):
            kwargs["table_name"] = self.data_structure.table

        # In the federated setting, the data should already have been loaded by the
        # Worker. So this call is effectively just for the local setting. The call is
        # idempotent so it doesn't matter if the data is already loaded.
        self.datasource.load_data(**kwargs)

    def _disallow_text_features(self) -> None:
        """Removes columns with semantic type TEXT from the data structure."""
        disallowed_columns = []
        for col in self.data_structure.selected_cols:
            if col in self.schema.get_feature_names(SemanticType.TEXT):
                disallowed_columns.append(col)
                logger.warning(
                    f"DataStructure has selected the text column {col} "
                    f"which is not supported. Removing this from the selection."
                )
        self.data_structure.ignore_cols = _add_this_to_list(
            disallowed_columns, self.data_structure.ignore_cols
        )
        self.data_structure.selected_cols = [
            i for i in self.data_structure.selected_cols if i not in disallowed_columns
        ]

    def _data_to_dataset(
        self,
        data_split: DataSplit,
    ) -> _BaseBitfountDataset:
        """Converts pandas dataframe to relevant BitfountDataset."""
        return self.data_factory.create_dataset(
            datasource=self.datasource,
            data_splitter=self.data_structure.data_splitter,
            target=self.target,
            schema=self.schema,
            selected_cols_semantic_types=self.data_structure.selected_cols_w_types,
            selected_cols=self.data_structure.selected_cols,
            batch_transforms=self.data_structure.get_batch_transformations(),
            data_split=data_split,
            weights_col=self.loss_weights_col,
            multihead_col=self.multihead_col,
            ignore_classes_col=self.ignore_classes_col,
        )

    def _create_datasets(self) -> None:
        """Creates datasets for dataloaders.

        Sets `self.train_ds`, `self.validation_ds` and `self.test_ds`
        """
        self.train_ds = self._data_to_dataset(DataSplit.TRAIN)
        self.validation_ds = self._data_to_dataset(DataSplit.VALIDATION)
        self.test_ds = self._data_to_dataset(DataSplit.TEST)

    def get_train_dataloader(
        self, batch_size: Optional[int] = None
    ) -> BitfountDataLoader:
        """Gets the relevant data loader for training data."""
        return self.data_factory.create_dataloader(self.train_ds, batch_size=batch_size)

    def get_validation_dataloader(
        self, batch_size: Optional[int] = None
    ) -> Optional[BitfountDataLoader]:
        """Gets the relevant data loader for validation data."""
        if not self.validation_ds:
            logging.warning(
                "No validation data in the dataset. Validation DataLoader is 'None'."
            )
            return None

        return self.data_factory.create_dataloader(
            self.validation_ds, batch_size=batch_size
        )

    def get_test_dataloader(
        self, batch_size: Optional[int] = None
    ) -> Optional[BitfountDataLoader]:
        """Gets the relevant data loader for test data."""
        if not self.test_ds:
            logging.warning("No test data in the dataset. Test DataLoader is 'None'.")
            return None

        return self.data_factory.create_dataloader(self.test_ds, batch_size=batch_size)



"""Manages the federated communication and training of models."""
from typing import List

from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.algorithms.column_avg import ColumnAverage
from bitfount.federated.algorithms.compute_intersection_rsa import (
    ComputeIntersectionRSA,
)
from bitfount.federated.algorithms.model_algorithms.evaluate import ModelEvaluation
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
)
from bitfount.federated.algorithms.model_algorithms.train_and_evaluate import (
    ModelTrainingAndEvaluation,
)
from bitfount.federated.algorithms.private_sql_query import PrivateSqlQuery
from bitfount.federated.algorithms.sql_query import SqlQuery
from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.early_stopping import FederatedEarlyStopping
from bitfount.federated.exceptions import (
    AggregatorError,
    BitfountTaskStartError,
    DecryptError,
    EncryptError,
    EncryptionError,
    MessageHandlerNotFoundError,
    MessageRetrievalError,
    PodConnectFailedError,
    PodNameError,
    PodRegistrationError,
    PrivateSqlError,
    SecureShareError,
)
from bitfount.federated.helper import combine_pod_schemas
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.modeller import _Modeller
from bitfount.federated.pod import Pod
from bitfount.federated.pod_keys_setup import PodKeys
from bitfount.federated.privacy.differential import DPModellerConfig, DPPodConfig
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
)
from bitfount.federated.protocols.psi import PrivateSetIntersection
from bitfount.federated.protocols.results_only import ResultsOnly
from bitfount.federated.roles import Role
from bitfount.federated.secure import SecureShare
from bitfount.federated.shim import BackendTensorShim
from bitfount.federated.transport import MAXIMUM_GRPC_MESSAGE_SIZE_BYTES
from bitfount.federated.transport.config import (
    PRODUCTION_MESSAGE_SERVICE_URL,
    MessageServiceConfig,
)
from bitfount.federated.types import AggregatorType, AlgorithmType, ProtocolType

__all__: List[str] = [
    "Aggregator",
    "AggregatorError",
    "AggregatorType",
    "AlgorithmType",
    "BackendTensorShim",
    "BitfountModelReference",
    "BitfountTaskStartError",
    "ColumnAverage",
    "DPModellerConfig",
    "DPPodConfig",
    "DecryptError",
    "EncryptError",
    "EncryptionError",
    "FederatedAveraging",
    "FederatedEarlyStopping",
    "FederatedModelTraining",
    "IdentityVerificationMethod",
    "MAXIMUM_GRPC_MESSAGE_SIZE_BYTES",
    "MessageHandlerNotFoundError",
    "MessageRetrievalError",
    "MessageServiceConfig",
    "ModelEvaluation",
    "ModelTrainingAndEvaluation",
    "_Modeller",
    "Pod",
    "PodConnectFailedError",
    "PodKeys",
    "PodNameError",
    "PodRegistrationError",
    "PrivateSetIntersection",
    "PrivateSqlError",
    "PrivateSqlQuery",
    "ProtocolType",
    "PRODUCTION_MESSAGE_SERVICE_URL",
    "ResultsOnly",
    "Role",
    "ComputeIntersectionRSA",
    "SecureAggregator",
    "SecureShare",
    "SecureShareError",
    "SqlQuery",
    "combine_pod_schemas",
]

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Contains courtesy classes and functions for making pod running easier."""
import importlib
import logging
import os
from os import PathLike
from typing import Union

import desert
from envyaml import EnvYAML

from bitfount.federated.pod import Pod
from bitfount.federated.pod_keys_setup import _get_pod_keys
from bitfount.hub.helper import _create_access_manager, _create_bitfounthub
from bitfount.runners.config_schemas import PodConfig

logger = logging.getLogger(__name__)


def setup_pod_from_config_file(path_to_config_yaml: Union[str, PathLike]) -> Pod:
    """Creates a pod from a YAML config file.

    Args:
        path_to_config_yaml: The path to the config file.

    Returns:
        The created pod.
    """
    logger.debug(f"Loading pod config from: {path_to_config_yaml}")

    # This double-underscore staticmethod access is not desirable but the additional
    # functionality in the class __init__ produces undesirable other inclusions.
    # This method allows us to just utilise the parsing aspect whilst performing
    # the envvar replacement.
    # TODO: [BIT-2202] Revisit double-underscore method usage.
    config_yaml = EnvYAML._EnvYAML__read_yaml_file(  # type: ignore[attr-defined] # Reason: see comment # noqa: B950
        path_to_config_yaml, os.environ, strict=True
    )
    config = desert.schema(PodConfig).load(config_yaml)
    return setup_pod_from_config(config)


def setup_pod_from_config(config: PodConfig) -> Pod:
    """Creates a pod from a loaded config.

    Args:
        config: The configuration as a PodConfig instance.

    Returns:
        The created pod.
    """
    bitfount_hub = _create_bitfounthub(config.username, config.hub.url)
    access_manager = _create_access_manager(
        bitfount_hub.session, config.access_manager.url
    )

    # Load Pod Keys
    pod_directory = bitfount_hub.user_storage_path / "pods" / config.name
    pod_keys = _get_pod_keys(pod_directory)

    try:
        datasource_cls = getattr(
            importlib.import_module("bitfount.data"), config.datasource
        )
    except AttributeError:
        raise ImportError(f"Unable to import {config.datasource} from bitfount.")
    datasource = datasource_cls(**config.data_config.datasource_args)

    return Pod(
        name=config.name,
        datasource=datasource,
        schema=config.schema,
        data_config=config.data_config,
        pod_details_config=config.pod_details_config,
        hub=bitfount_hub,
        message_service=config.message_service,
        access_manager=access_manager,
        pod_keys=pod_keys,
        approved_pods=config.approved_pods,
        differential_privacy=config.differential_privacy,
    )



"""Pods for responding to tasks."""
import asyncio
from contextlib import nullcontext
from dataclasses import asdict
import os
from pathlib import Path
import threading
from typing import (
    Callable,
    ContextManager,
    Coroutine,
    Dict,
    Iterable,
    List,
    MutableSequence,
    Optional,
    Tuple,
    Union,
    cast,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from requests import HTTPError, RequestException
import yaml

from bitfount.config import BITFOUNT_STORAGE_PATH
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.exceptions import DataSourceError
from bitfount.data.schema import BitfountSchema
from bitfount.federated.aggregators.secure import _is_secure_share_task_request
from bitfount.federated.authorisation_checkers import (
    _IDENTITY_VERIFICATION_METHODS_MAP,
    IdentityVerificationMethod,
    _AuthorisationChecker,
    _OIDCAuthorisationCode,
    _OIDCDeviceCode,
    _SAMLAuthorisation,
    _SignatureBasedAuthorisation,
    check_identity_verification_method,
)
from bitfount.federated.exceptions import PodNameError, PodRegistrationError
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_and_connect_pod_mailbox,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.monitoring import task_monitor_context
from bitfount.federated.pod_keys_setup import PodKeys, _get_pod_keys
from bitfount.federated.pod_response_message import _PodResponseMessage
from bitfount.federated.pod_vitals import _PodVitals, _PodVitalsHandler
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.task_requests import (
    _SignedEncryptedTaskRequest,
    _TaskRequestMessage,
)
from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
)
from bitfount.federated.transport.pod_transport import _PodMailbox
from bitfount.federated.transport.worker_transport import (
    _InterPodWorkerMailbox,
    _WorkerMailbox,
)
from bitfount.federated.types import (
    AggregatorType,
    SerializedProtocol,
    _PodResponseType,
)
from bitfount.federated.utils import _keys_exist
from bitfount.federated.worker import _Worker
from bitfount.hub.api import BitfountAM, BitfountHub, PodPublicMetadata
from bitfount.hub.authentication_flow import _DEFAULT_USERNAME, _get_auth_environment
from bitfount.hub.exceptions import SchemaUploadError
from bitfount.hub.helper import (
    _create_access_manager,
    _create_bitfounthub,
    _get_pod_public_keys,
)
from bitfount.runners.config_schemas import (
    POD_NAME_REGEX,
    AccessManagerConfig,
    PodConfig,
    PodDataConfig,
    PodDetailsConfig,
)
from bitfount.transformations.dataset_operations import (
    CleanDataTransformation,
    NormalizeDataTransformation,
)
from bitfount.transformations.processor import TransformationProcessor
from bitfount.utils import _handle_fatal_error, _is_notebook

logger = _get_federated_logger(__name__)


class Pod:
    """Makes data and computation available remotely and responds to tasks.

    The basic component of the Bitfount network is the `Pod` (Processor of Data). Pods
    are co-located with data, check users are authorized to do given operations on the
    data and then do any approved computation. Creating a `Pod` will register the pod
    with Bitfount Hub.

    ```python title="Example usage:"
    import bitfount as bf

    pod = bf.Pod(
        name="really_cool_data",
        data="/path/to/data",
    )
    pod.start()
    ```

    :::tip

    Once you start a `Pod`, you can just leave it running in the background. It will
    automatically respond to any tasks without any intervention required.

    :::

    Args:
        name: Name of the pod. This will appear on `Bitfount Hub` and `Bitfount AM`.
            This is also used for the name of the table in a single-table `BaseSource`.
        datasource: A concrete instance of the `BaseSource` object.
        username: Username of the user who is registering the pod. Defaults to None.
        data_config: Configuration for the data. Defaults to None.
        schema: Schema for the data. This can be a `BitfountSchema` object or a Path
            to a serialized `BitfountSchema`. This will generated automatically if not
            provided. Defaults to None.
        pod_details_config: Configuration for the pod details. Defaults to None.
        bitfount_hub: Bitfount Hub to register the pod with. Defaults to None.
        ms_config: Configuration for the message service. Defaults to None.
        access_manager: Access manager to use for checking access. Defaults to None.
        pod_keys: Keys for the pod. Defaults to None.
        approved_pods: List of other pod identifiers this pod is happy
            to share a training task with. Required if the protocol uses the
            `SecureAggregator` aggregator.
        pod_dp: Differential privacy configuration for the pod. Defaults to None.

    Attributes:
        data: Data that the pod encapsulates.
        name: Name of the pod.
        pod_identifier: Identifier of the pod.
        private_key: Private key of the pod.
        public_key: Public key of the pod.
        public_metadata: Public metadata about the pod.
        schema: Schema for the data.

    Raises:
        PodRegistrationError: If the pod could not be registered for any reason.
        DataSourceError: If the `BaseSource` for the provided datasource has
            not been initialised properly. This can be done by calling
            `super().__init__(**kwargs)` in the `__init__` of the DataSource.
    """

    def __init__(
        self,
        name: str,
        datasource: BaseSource,
        username: Optional[str] = None,
        data_config: Optional[PodDataConfig] = None,
        schema: Optional[Union[str, os.PathLike, BitfountSchema]] = None,
        pod_details_config: Optional[PodDetailsConfig] = None,
        hub: Optional[BitfountHub] = None,
        message_service: Optional[MessageServiceConfig] = None,
        access_manager: Optional[BitfountAM] = None,
        pod_keys: Optional[PodKeys] = None,
        approved_pods: Optional[List[str]] = None,
        differential_privacy: Optional[DPPodConfig] = None,
    ):
        self.name = name

        if not data_config:
            data_config = PodDataConfig()

        pod_details_config = (
            pod_details_config
            if pod_details_config is not None
            else self._get_default_pod_details_config()
        )
        self._pod_config = PodConfig(
            datasource=datasource.__class__.__name__,
            data_config=data_config,
            pod_details_config=pod_details_config,
            name=name,
            schema=schema if isinstance(schema, Path) else None,
            access_manager=AccessManagerConfig(),
            message_service=message_service
            if message_service
            else MessageServiceConfig(),
            approved_pods=approved_pods,
        )
        # Fast-fail if BaseSource has not been initialised
        if not datasource.is_initialised:
            raise DataSourceError(
                "The datasource provided has not initialised the BaseSource "
                "parent class. Please make sure that you call "
                "`super().__init__(**kwargs)` in your child method."
            )
        # Create datasource
        self.datasource = datasource

        # Load schema if necessary
        if schema and not isinstance(schema, BitfountSchema):
            schema = BitfountSchema.load_from_file(schema)

        # If the schema was a path, it should be loaded by here, so we can cast:
        schema = cast(Optional[BitfountSchema], schema)
        self.schema = self._setup_schema(schema=schema, data_config=data_config)

        self.public_metadata = self._get_public_metadata(
            pod_details_config, self.schema
        )

        self._hub = hub if hub is not None else _create_bitfounthub(username=username)
        self._session = self._hub.session

        self._access_manager = (
            access_manager
            if access_manager is not None
            else _create_access_manager(self._session)
        )
        self._access_manager_public_key = self._access_manager.get_access_manager_key()

        self.private_key, self.pod_public_key = self._get_default_pod_keys(pod_keys)

        self.pod_identifier = f"{self._session.username}/{self.name}"
        if approved_pods is None:
            approved_pods = []
        approved_pods = _check_and_update_pod_ids(
            [*approved_pods, self.pod_identifier], self._hub
        )
        self.approved_pods = approved_pods
        self._pod_dp = differential_privacy
        self._pod_vitals = _PodVitals()

        # Connecting the pod to the message service must happen AFTER registering
        # it on the hub as the message service uses hub information to verify that
        # the relevant message queue is available.
        try:
            self._register_pod()
        except PodRegistrationError as pre:
            _handle_fatal_error(pre, logger=logger)

        self._ms_config: Optional[MessageServiceConfig] = message_service
        self._mailbox: Optional[_PodMailbox] = None

        # Marker for when initialization is complete
        self._initialised: bool = False

    @property
    def name(self) -> str:
        """Pod name property."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Validate Pod's name matches POD_NAME_REGEX."""
        if _name := POD_NAME_REGEX.fullmatch(name):
            self._name = _name.string
        else:
            raise PodNameError(
                f"Invalid Pod name: {name}. "
                f"Pod names must match: {POD_NAME_REGEX.pattern}"
            )

    def to_yaml(self, output_file: Optional[str] = None) -> Optional[Dict]:
        """Convert the pod specification to yaml.

        If output_file argument is given, write yaml to that locaton.
        """
        output = asdict(self._pod_config)
        if output_file:
            with open(output_file, "w") as file:
                yaml.dump(output, file)
            return None
        else:
            return output

    def _setup_schema(
        self,
        data_config: PodDataConfig,
        schema: Optional[BitfountSchema] = None,
    ) -> BitfountSchema:
        """Generate pod schema."""
        self.datasource.load_data()
        logger.info("Generating the schema ...")
        # Create schema if not provided
        if not schema:
            schema = BitfountSchema()

        # Add BaseSource to schema
        schema.add_datasource_tables(
            datasource=self.datasource,
            table_name=self.name,
            ignore_cols=data_config.ignore_cols,
            force_stypes=data_config.force_stypes,
        )

        # Auto-tidy if specified and datasource is not multi-table
        if data_config.auto_tidy and self.datasource.multi_table:
            logger.warning("Can't autotidy multi-table data.")
        elif data_config.auto_tidy:
            clean_data = CleanDataTransformation()
            # Normalization is applied to all float columns if `auto-tidy` is true
            normalize = NormalizeDataTransformation()
            processor = TransformationProcessor(
                [clean_data, normalize],
                schema.get_table_schema(self.name),
            )
            self.datasource.data = processor.transform(self.datasource.data)

            # Add BaseSource to schema again because features will have changed by
            # auto-tidying
            schema.add_datasource_tables(
                datasource=self.datasource,
                table_name=self.name,
                ignore_cols=data_config.ignore_cols,
                force_stypes=data_config.force_stypes,
            )

        # Freeze schema
        schema.freeze()
        return schema

    def _get_default_pod_details_config(self) -> PodDetailsConfig:
        """Get default pod details config."""
        return PodDetailsConfig(display_name=self.name, description=self.name)

    def _get_public_metadata(
        self, pod_details_config: PodDetailsConfig, schema: BitfountSchema
    ) -> PodPublicMetadata:
        """Get PodPublicMetadata."""
        return PodPublicMetadata(
            self.name,
            pod_details_config.display_name,
            pod_details_config.description,
            schema.to_json(),
        )

    def _get_default_pod_keys(
        self, pod_keys: Optional[PodKeys]
    ) -> Tuple[RSAPrivateKey, RSAPublicKey]:
        """Get default pod keys."""
        if pod_keys is None:
            user_storage_path = BITFOUNT_STORAGE_PATH / _DEFAULT_USERNAME
            pod_directory = user_storage_path / "pods" / self.name
            pod_keys = _get_pod_keys(pod_directory)
        return pod_keys.private, pod_keys.public

    def _register_pod(self) -> None:
        """Register pod with Bitfount Hub.

        If Pod is already registered, will update pod details if anything has changed.

        Raises:
            PodRegistrationError: if registration fails for any reason
        """
        try:
            logger.info("Registering/Updating details on Bitfount Hub.")
            self._hub.register_pod(
                self.public_metadata,
                self.pod_public_key,
                self._access_manager_public_key,
            )
        except (HTTPError, SchemaUploadError) as ex:
            logger.critical(f"Failed to register with hub: {ex}")
            raise PodRegistrationError("Failed to register with hub") from ex
        except RequestException as ex:
            logger.critical(f"Could not connect to hub: {ex}")
            raise PodRegistrationError("Could not connect to hub") from ex

    async def _initialise(self) -> None:
        """Initialises the pod.

        Sets any attributes that could not be created at creation time.
        """
        if not self._initialised:
            # Create mailbox. Cannot be done in __init__ due to async nature.
            self._mailbox = await _create_and_connect_pod_mailbox(
                pod_name=self.name, session=self._session, ms_config=self._ms_config
            )

            # Set initialised state
            self._initialised = True
        else:
            logger.warning("Pod._initialise() called twice. This is not allowed.")

    def _secure_aggregation_other_workers_response(
        self, other_worker_names: MutableSequence[str]
    ) -> Optional[List[str]]:
        """Checks if secure aggregation can be performed with given other workers.

        Args:
            other_worker_names (List[str]): list of other worker names

        Returns:
            Optional[List[str]]:
                unapproved workers (if they exist in other_worker_names)
        """
        unapproved_pods = [
            worker for worker in other_worker_names if worker not in self.approved_pods
        ]
        logger.debug(
            f"Modeller requested aggregation"
            f" with non-approved pods: {unapproved_pods}"
        )

        if unapproved_pods:
            logger.info(
                "Modeller requested aggregation with"
                " pods that this pod has not approved."
            )
            return unapproved_pods

        logger.debug("All pods requested by modeller for aggregation are approved.")
        return None

    def _check_for_unapproved_pods(
        self,
        pods_involved_in_task: Iterable[str],
        serialized_protocol: SerializedProtocol,
    ) -> Optional[List[str]]:
        """Returns the pods that we're not happy to work with.

        If secure aggregation has been requested then this will
        identify any pods that we've not approved.

        In any other case it returns None, as there's no concern
        around security with other pods.

        Args:
            pods_involved_in_task: A list of other pods that have been contacted by
                the modeller for this task.
            serialized_protocol: The decrypted serialized protocol portion of the task
                request.

        Returns:
            Either a list of unapproved pods or `None` if all are approved or if secure
            aggregation not in use.
        """
        unapproved_workers = None

        # Create mutable version of pods_involved_in_task
        other_pods: List[str] = list(pods_involved_in_task)

        # We don't need to check if we're approved to work with our self.
        try:
            other_pods.remove(self.pod_identifier)
        except ValueError:  # if not in list to remove
            pass

        aggregator = serialized_protocol.get("aggregator")
        if (
            aggregator
            and aggregator["class_name"] == AggregatorType.SecureAggregator.value
        ):
            logger.info(
                "Secure aggregation is in use, checking responses from other pods."
            )
            unapproved_workers = self._secure_aggregation_other_workers_response(
                other_pods
            )

        return unapproved_workers

    async def _new_task_request_handler(self, message: _BitfountMessage) -> None:
        """Called on new task request being received from message service."""
        logger.info(f"Training task request received from '{message.sender}'")
        try:
            await self._create_and_run_worker(message)
        except asyncio.TimeoutError:
            logger.info("Ready for next task...")
            return

    async def _create_and_run_worker(self, message: _BitfountMessage) -> None:
        """Creates and runs a worker instance."""
        # `_initialise` is always called before this method, so we can assume
        # that the mailbox is initialised. Reassuring mypy that this is True.
        assert isinstance(self._mailbox, _PodMailbox)  # nosec[assert_used]

        # Unpack task details from received message
        logger.info("Unpacking task details from message...")
        task_id = message.task_id
        task_request_message: _TaskRequestMessage = _TaskRequestMessage.deserialize(
            message.body
        )
        auth_type: IdentityVerificationMethod = check_identity_verification_method(
            task_request_message.auth_type
        )
        authoriser_cls = _IDENTITY_VERIFICATION_METHODS_MAP[auth_type]
        task_request = authoriser_cls.unpack_task_request(
            message.body, self.private_key
        )

        # If we are using secure aggregation we check for unapproved workers; if
        # we are not, `unapproved_workers` will be `None`.
        other_pods = [
            pod_id
            for pod_id in message.pod_mailbox_ids
            if pod_id != self.pod_identifier
        ]
        unapproved_workers = self._check_for_unapproved_pods(
            other_pods, task_request.serialized_protocol
        )

        # If we are dealing with secure aggregation (and hence need inter-pod
        # communication) we create an appropriate mailbox as long as there are no
        # unapproved workers.
        # If there are, the task will be rejected, so we can just create a normal
        # mailbox (as don't need inter-pod communication to reject the task).
        # Similarly, if we're not using secure aggregation we just create a normal
        # mailbox as inter-pod communication won't be needed.
        worker_mailbox: _WorkerMailbox
        if _is_secure_share_task_request(task_request) and not unapproved_workers:
            logger.debug("Creating mailbox with inter-pod support.")

            other_pod_public_keys = _get_pod_public_keys(other_pods, self._hub)

            worker_mailbox = _InterPodWorkerMailbox(
                pod_public_keys=other_pod_public_keys,
                private_key=self.private_key,
                pod_identifier=self.pod_identifier,
                modeller_mailbox_id=message.sender_mailbox_id,
                modeller_name=message.sender,
                aes_encryption_key=task_request.aes_key,
                message_service=self._mailbox.message_service,
                pod_mailbox_ids=message.pod_mailbox_ids,
                task_id=task_id,
            )
        else:
            logger.debug("Creating modeller<->worker-only mailbox.")
            worker_mailbox = _WorkerMailbox(
                pod_identifier=self.pod_identifier,
                modeller_mailbox_id=message.sender_mailbox_id,
                modeller_name=message.sender,
                aes_encryption_key=task_request.aes_key,
                message_service=self._mailbox.message_service,
                pod_mailbox_ids=message.pod_mailbox_ids,
                task_id=task_id,
            )

        # TODO: [BIT-1045] Move the secure aggregation allowed check to the access
        #       manager once we support configuring or storing it there.
        if unapproved_workers:
            # There are pods we're explicitly not happy to work with (i.e. we're
            # using secure aggregation) we reject the task.
            logger.info(f"Task from '{message.sender}' rejected.")
            authorisation_errors = _PodResponseMessage(
                message.sender, self.pod_identifier
            )
            authorisation_errors.add(
                _PodResponseType.NO_ACCESS,
                unapproved_workers,
            )
            await worker_mailbox.reject_task(authorisation_errors.messages)
            return

        logger.debug("Creating authorisation checker.")
        authorisation_checker = self._create_authorisation_checker(
            task_request_message=task_request_message,
            sender=message.sender,
            worker_mailbox=worker_mailbox,
        )

        logger.debug("Creating worker.")
        worker = _Worker(
            datasource=self.datasource,
            mailbox=worker_mailbox,
            bitfounthub=self._hub,
            authorisation=authorisation_checker,
            pod_vitals=self._pod_vitals,
            pod_dp=self._pod_dp,
            pod_identifier=self.pod_identifier,
            serialized_protocol=task_request.serialized_protocol,
        )

        # If interacting with an older modeller version then task_id won't be supplied
        task_monitor_cm: ContextManager
        if worker_mailbox.task_id:
            task_monitor_cm = task_monitor_context(
                hub=self._hub,
                task_id=worker_mailbox.task_id,
                sender_id=worker_mailbox.mailbox_id,
            )
        else:
            task_monitor_cm = nullcontext()

        with task_monitor_cm:
            # Run the worker and the mailbox listening simultaneously
            try:
                await _run_func_and_listen_to_mailbox(worker.run(), worker_mailbox)
            except Exception as e:
                logger.federated_error(e, exc_info=True)
                logger.exception(e)

                if worker_mailbox.task_id:
                    logger.error(
                        f"Exception whilst running task {worker_mailbox.task_id}."
                    )
                else:
                    logger.error("Exception whilst running task.")

        logger.info("Ready for next task...")

    def _create_authorisation_checker(
        self,
        task_request_message: _TaskRequestMessage,
        sender: str,
        worker_mailbox: _WorkerMailbox,
    ) -> _AuthorisationChecker:
        """Create appropriate Authorisation Checker.

        Determines checker to create based on supplied auth_type.

        Args:
            task_request_message: The full task request message.
            sender: The sender (i.e. modeller) of the request.
            worker_mailbox: Worker mailbox for communication with modeller.

        Returns:
            An authorisation checker.
        """
        auth_type: IdentityVerificationMethod = check_identity_verification_method(
            task_request_message.auth_type
        )
        authorisation_checker_cls = _IDENTITY_VERIFICATION_METHODS_MAP[auth_type]

        task_request = authorisation_checker_cls.unpack_task_request(
            task_request_message, self.private_key
        )
        serialized_protocol = task_request.serialized_protocol
        # remove schema since it is the largest task element.
        if _keys_exist(serialized_protocol, "algorithm", "model", "schema"):  # type: ignore[arg-type] # Reason: a typed dict is still a dict. # noqa: B950
            serialized_protocol["algorithm"]["model"].pop("schema")  # type: ignore[typeddict-item] # Reason: all models have a schema associated with them # noqa: B950

        pod_response_message = _PodResponseMessage(
            modeller_name=sender,
            pod_identifier=self.pod_identifier,
        )

        authorisation_checker: _AuthorisationChecker

        if auth_type == IdentityVerificationMethod.KEYS:
            # Public Key Signature authorisation
            packed_request: _SignedEncryptedTaskRequest = (
                authorisation_checker_cls.extract_from_task_request_message(
                    task_request_message
                )
            )

            authorisation_checker = _SignatureBasedAuthorisation(
                pod_response_message=pod_response_message,
                access_manager=self._access_manager,
                modeller_name=worker_mailbox.modeller_name,
                encrypted_task_request=packed_request.encrypted_request,
                signature=packed_request.signature,
                serialized_protocol=serialized_protocol,
            )
        elif auth_type == IdentityVerificationMethod.OIDC_ACF_PKCE:
            # OIDC Authorization Code Flow
            auth_env = _get_auth_environment()
            authorisation_checker = _OIDCAuthorisationCode(
                pod_response_message=pod_response_message,
                access_manager=self._access_manager,
                mailbox=worker_mailbox,
                serialized_protocol=serialized_protocol,
                _auth_domain=auth_env.auth_domain,
                _client_id=auth_env.client_id,
            )
        elif auth_type == IdentityVerificationMethod.OIDC_DEVICE_CODE:
            # OIDC Device Code flow
            auth_env = _get_auth_environment()
            authorisation_checker = _OIDCDeviceCode(
                pod_response_message=pod_response_message,
                access_manager=self._access_manager,
                mailbox=worker_mailbox,
                serialized_protocol=serialized_protocol,
                _auth_domain=auth_env.auth_domain,
                _client_id=auth_env.client_id,
            )
        else:
            # Default to SAML Authorisation
            authorisation_checker = _SAMLAuthorisation(
                pod_response_message=pod_response_message,
                access_manager=self._access_manager,
                mailbox=worker_mailbox,
                serialized_protocol=serialized_protocol,
            )
        return authorisation_checker

    @staticmethod
    async def _repeat(interval: int, func: Callable[..., Coroutine]) -> None:
        """Run coroutine func every interval seconds.

        If func has not finished before *interval*, will run again
        immediately when the previous iteration finished.

        Args:
            interval: run interval in seconds
            func: function to call which returns a coroutine to await
        """
        while True:
            await asyncio.gather(func(), asyncio.sleep(interval))

    async def _pod_heartbeat(self) -> None:
        """Makes a pod heartbeat to the hub."""
        try:
            self._hub.do_pod_heartbeat(self.name, self.pod_public_key)
        except HTTPError as ex:
            logger.warning(f"Failed to reach hub for status: {ex}")
        except RequestException as ex:
            logger.warning(f"Could not connect to hub for status: {ex}")

    def _run_pod_heartbeat_task(self) -> None:
        """Makes 10-second interval pod heartbeats to the hub."""
        if _is_notebook():
            # We need to create a new event loop here for jupyter
            # As it's run in a new thread and can't be patched by nest_asyncio
            asyncio.set_event_loop(asyncio.new_event_loop())
        asyncio.run(self._repeat(10, self._pod_heartbeat))

    def _pod_vitals_server(self, vitals_handler: _PodVitalsHandler) -> None:
        """Run _PodVitals webserver."""
        # The Pod Vitals webserver should run until the
        # pod itself it shut down. asyncio.run would handle
        # the event loop for us however it would also
        # shutdown the loop (and the webserver) on completion
        # so instead we directly interact with the
        # event loop here to ensure it is run_forever.
        pod_vitals_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(pod_vitals_loop)
        vitals_handler.start(pod_vitals_loop)
        pod_vitals_loop.run_forever()

    def _run_pod_vitals_server(self) -> Optional[_PodVitalsHandler]:
        """Create _PodVitalsHandelr and run _PodVitals webserver."""
        # Check that we have not initialized the Pod from a notebook
        if not _is_notebook():
            # Setup pod vitals webserver
            vitals_handler = _PodVitalsHandler(self._pod_vitals)
            logger.debug("Starting Pod Vitals interface...")
            threading.Thread(
                daemon=True, target=self._pod_vitals_server, args=(vitals_handler,)
            ).start()
            return vitals_handler
        else:
            return None

    async def start_async(self) -> None:
        """Starts a pod instance, listening for tasks.

        Whenever a task is received, a worker is created to handle it. Runs continuously
        and asynchronously orchestrates training whenever a task arrives i.e. multiple
        tasks can run concurrently.
        """
        # Do post-init initialization work
        await self._initialise()

        # `_initialise` has just been called which sets the mailbox so we can assume
        # that the mailbox is initialised. Reassuring mypy that this is True.
        assert isinstance(self._mailbox, _PodMailbox)  # nosec[assert_used]

        # Setup heartbeat to hub
        logger.info(f"Starting pod {self.name}...")
        threading.Thread(daemon=True, target=self._run_pod_heartbeat_task).start()

        # Start pod vitals webserver
        vitals_handler = self._run_pod_vitals_server()

        # Attach handler for new tasks
        self._mailbox.register_handler(
            _BitfountMessageType.JOB_REQUEST, self._new_task_request_handler
        )

        # Start pod listening for messages
        logger.info("Pod started... press Ctrl+C to stop")
        await self._mailbox.listen_indefinitely()
        logger.info(f"Pod {self.name} stopped.")

        # Shutdown pod vitals webserver
        if vitals_handler:
            await vitals_handler.runner.cleanup()

    def start(self) -> None:
        """Starts a pod instance, listening for tasks.

        Whenever a task is received, a worker is created to handle it. Runs continuously
        and asynchronously orchestrates training whenever a task arrives i.e. multiple
        tasks can run concurrently.
        """
        asyncio.run(self.start_async())



"""Parameter clipping for continuous attributes secure aggergation.

Based on discussions from:
https://discuss.pytorch.org/t/set-constraints-on-parameters-or-layers/23620
"""
import logging

import torch
import torch.nn as nn

logger = logging.getLogger(__name__)


class _PytorchParamConstraint:
    """Class for clamping model parameters.

    Used to make sure that the model parameters stay within
    the constraints for secure aggregation.
    """

    def __init__(self, prime_q: int, precision: int, num_workers: int) -> None:
        self.prime_q = prime_q
        self.precision = precision
        self.num_workers = num_workers

    def __call__(self, module: nn.Module) -> None:
        """Clamp the required parameters on call."""
        if hasattr(module, "weight"):
            # Get the parameter data
            w = module.weight.data
            # Clamp the data in the required interval
            w = torch.clamp(
                w,
                -self.prime_q / (self.precision * 2 * self.num_workers),
                self.prime_q / (self.precision * 2 * self.num_workers),
            )
            # Assign the clamped value to the parameter
            module.weight.data = w

        if hasattr(module, "running_var"):
            # Get the parameter data
            w = module.running_var.data
            # Clamp the data in the required interval
            w = torch.clamp(
                w,
                -self.prime_q / (self.precision * 2 * self.num_workers),
                self.prime_q / (self.precision * 2 * self.num_workers),
            )
            # Assign the clamped value to the parameter
            module.running_var.data = w

        if hasattr(module, "running_mean"):
            # Get the parameter data
            w = module.running_mean.data
            # Clamp the data in the required interval
            w = torch.clamp(
                w,
                -self.prime_q / (self.precision * 2 * self.num_workers),
                self.prime_q / (self.precision * 2 * self.num_workers),
            )
            # Assign the clamped value to the parameter
            module.running_mean.data = w



"""Helper module for federated transport."""
import asyncio
from functools import wraps
import logging
from statistics import mean
import threading
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Final,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

from grpc import RpcError, StatusCode
from typing_extensions import Concatenate, ParamSpec

from bitfount.federated.transport.exceptions import BitfountMessageServiceError

_logger = logging.getLogger(__name__)


def _average_training_metrics(
    validation_metrics: Sequence[Mapping[str, str]]
) -> Dict[str, float]:
    """Average training metrics from each worker."""
    averaged_metrics = dict()
    if validation_metrics:
        # What should happen if one (or all) of the pods does not respond in time?
        for metric_key in validation_metrics[0]:
            averaged_metrics[metric_key] = mean(
                float(worker_metrics[metric_key])
                for worker_metrics in validation_metrics
            )
    return averaged_metrics


_RETRY_STATUS_CODES: Final = {
    sc
    for sc in StatusCode
    if sc
    not in (
        # because this means it worked
        StatusCode.OK,
        # because this means it doesn't need to be done again
        StatusCode.ALREADY_EXISTS,
        # because this means it will never work
        StatusCode.UNIMPLEMENTED,
    )
}
_DEFAULT_TIMEOUT: Final = 20.0
_DEFAULT_MAX_RETRIES: Final = 3
_DEFAULT_BACKOFF_FACTOR: Final = 1

# These should be replaced with ParamSpec versions once
# https://github.com/python/mypy/issues/11855 is resolved
_F = TypeVar("_F", bound=Callable[..., Awaitable[Any]])


@overload
def _auto_retry_grpc(
    original_rpc_func: _F,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
    additional_no_retry_status_codes: Optional[Collection[StatusCode]] = None,
) -> _F:
    """Applies automatic retries to gRPC calls when encountering specific errors."""
    ...


@overload
def _auto_retry_grpc(
    original_rpc_func: None = None,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
    additional_no_retry_status_codes: Optional[Collection[StatusCode]] = None,
) -> Callable[[_F], _F]:
    """Applies automatic retries to gRPC calls when encountering specific errors."""
    ...


def _auto_retry_grpc(
    original_rpc_func: Optional[_F] = None,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
    additional_no_retry_status_codes: Optional[Collection[StatusCode]] = None,
) -> Union[_F, Callable[[_F], _F]]:
    """Applies automatic retries to gRPC calls when encountering specific errors.

    Wraps the target gRPC call in a retry mechanism which will reattempt
    the call if a retryable gRPC error response is received.

    Utilises an exponential backoff to avoid flooding the request and to give
    time for the issue to resolve itself.

    Can be used as either an argumentless decorator (@_auto_retry_grpc) or a
    decorator with args (@_auto_retry_grpc(...)).
    """
    if additional_no_retry_status_codes:
        _additional_no_retry_status_codes = set(additional_no_retry_status_codes)
    else:
        _additional_no_retry_status_codes = set()

    def _decorate(grpc_func: _F) -> _F:
        """Apply decoration to target request function."""

        @wraps(grpc_func)
        async def _wrapped_async_grpc_func(*args: Any, **kwargs: Any) -> Any:
            """Wraps target gRPC function in retry capability.

            Adds automatic retry, backoff, and logging.
            """
            # Set default timeout if one not provided
            timeout = kwargs.get("timeout", None)
            if timeout is None:
                logging.debug(
                    f"No gRPC timeout provided,"
                    f" setting to default timeout ({_DEFAULT_TIMEOUT}s)"
                )
                kwargs["timeout"] = _DEFAULT_TIMEOUT

            retry_count = 0

            while retry_count <= max_retries:
                final_retry = retry_count == max_retries

                # Attempt to make wrapped call and handle if it doesn't work
                # as expected
                try:
                    return await grpc_func(*args, **kwargs)
                except RpcError as ex:
                    # If an error occurs, we can retry unless this is our final
                    # attempt, or the error code is a non-retryable one.
                    if (
                        final_retry
                        or ex.code() not in _RETRY_STATUS_CODES
                        or ex.code() in _additional_no_retry_status_codes
                    ):
                        raise ex
                    else:
                        failure_cause_msg = (
                            f"gRPC error occurred:"
                            f" (Status Code {ex.code()}) {ex.details()}"
                        )

                # If we reach this point we must be attempting a retry
                retry_count += 1

                # Backoff is increased using standard exponential backoff formula
                # For standard backoff factor of one this results in backoffs of
                # [1, 2, 4, 8, ...] seconds.
                backoff = backoff_factor * (2 ** (retry_count - 1))

                # Log out failure information and retry information.
                _logger.debug(
                    f"{failure_cause_msg}; "
                    f"will retry in {backoff} seconds (attempt {retry_count})."
                )

                await asyncio.sleep(backoff)

            # We shouldn't reach this point due to how the loop can be exited,
            # but just in case.
            raise BitfountMessageServiceError(
                "Unable to make connection, even after multiple attempts."
            )

        return cast(_F, _wrapped_async_grpc_func)

    if original_rpc_func:
        # Was used as @_auto_retry_grpc (or called directly).
        # original_rpc_func was passed in through the decorator machinery so just
        # wrap and return.
        return _decorate(original_rpc_func)
    else:
        # Was used as @_auto_retry_grpc(**kwargs).
        # original_rpc_func not yet passed in so need to return a decorator function
        # to allow the decorator machinery to pass it in.
        return _decorate


# Decorator for thread-safe synchronisation of methods
_LOCK = TypeVar("_LOCK", bound=Union[threading.Lock, threading.RLock])


class _Synchronisable(Protocol[_LOCK]):
    """Represents a class that can be synchronised."""

    _sync_lock: _LOCK


Self = TypeVar(
    "Self",
    bound=Union[_Synchronisable[threading.Lock], _Synchronisable[threading.RLock]],
)
_P = ParamSpec("_P")
_R = TypeVar("_R")


def _synchronised(
    meth: Callable[Concatenate[Self, _P], _R]
) -> Callable[Concatenate[Self, _P], _R]:
    """Wrap target method with instance's Lock."""

    @wraps(meth)
    def _wrapper(self_: Self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        _logger.debug(
            f"Waiting on {self_.__class__.__name__} lock in"
            f" {threading.current_thread().name}"
        )

        with self_._sync_lock:
            _logger.debug(
                f"Got {self_.__class__.__name__} lock in"
                f" {threading.current_thread().name}"
            )
            res = meth(self_, *args, **kwargs)

        _logger.debug(
            f"Released {self_.__class__.__name__} lock in"
            f" {threading.current_thread().name}"
        )
        return res

    return _wrapper



"""Custom exceptions for the federated package."""
from bitfount.exceptions import BitfountError


class BitfountTaskStartError(BitfountError, RuntimeError):
    """Raised when an issue occurs whilst trying to start a task with pods."""

    pass


class MessageHandlerNotFoundError(BitfountError, KeyError):
    """Error raised when registered message handler can't be found."""

    pass


class MessageRetrievalError(BitfountError, RuntimeError):
    """Raised when an error occurs whilst retrieving a message from message service."""

    pass


class PodConnectFailedError(BitfountError, TypeError):
    """The message service has not correctly connected the pod."""

    pass


class PodRegistrationError(BitfountError):
    """Error related to registering a Pod with BitfountHub."""

    pass


class PodResponseError(BitfountError):
    """Pod rejected or failed to respond to a task request."""

    pass


class PodNameError(BitfountError):
    """Error related to given Pod name."""

    pass


class PrivateSqlError(BitfountError):
    """An exception for any issues relating to the PrivateSQL algorithm."""

    pass


class SecureShareError(BitfountError):
    """Error related to SecureShare processes."""

    pass


class AggregatorError(BitfountError, ValueError):
    """Error related to Aggregator classes."""

    pass


class EncryptionError(BitfountError):
    """Error related to encryption processes."""

    pass


class EncryptError(EncryptionError):
    """Error when attempting to encrypt."""

    pass


class DecryptError(EncryptionError):
    """Error when attempting to decrypt."""

    pass


class DPParameterError(BitfountError):
    """Error if any of given dp params are not allowed."""

    pass


class PSIError(BitfountError):
    """Error related to the PrivateSetIntersection protocol."""

    pass


class BlindingError(PSIError):
    """Error when attempting to blind."""

    pass


class UnBlindingError(PSIError):
    """Error when attempting to unblind."""

    pass


class OutOfBoundsError(PSIError):
    """Error when a value is out of bounds."""

    pass


class PSINoDataSourceError(PSIError):
    """Error when modeller tries to run a PSI task without a datasource."""

    pass


class PSIMultiplePodsError(PSIError):
    """Error when modeller tries to run a PSI task on multiple pods."""

    pass


class PSIMultiTableError(PSIError):
    """Error when trying perform PSI on a multitable datasource without specifying a table name."""  # noqa: B950

    pass
